/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#ifdef __USAGE		
%C [options]

Options 
	-V      Verbose mode
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <stddef.h>
#include <signal.h>
#include <time.h>
#include <stdarg.h>

static int Qargc;
static char **Qargv;
static char **Qenvp;

static char	MyURL[256];
static int verbose;

#define	ImageMod	4
#define	ColourMod	7
#define	BGColourMod	7
#define	TotalMod	(ImageMod * ColourMod * BGColourMod)

#define	ColourDiv	(ImageMod)
#define	BGColourDiv	(ColourDiv*ColourMod)

typedef unsigned long ulong_t;
typedef unsigned char uchar_t;


static ulong_t Colours[] = {
	0xFF0000,		// red
	0xFF8000,		// orange
	0xFFFF00,		// yellow
	0x00FF00,		// green
	0x0000FF,		// blue
	0xCC00FF,		// purple
	0xFFFFFF,
	0xFFFFFF,
	0xFFFFFF,
	0xFFFFFF,
};

static ulong_t BGColours[] = {
	0x000000,		// black
	0xAA0010,		// red
	0x008000,		// green
	0x0000AA,		// blue
	0x8000AA,		// purple
	0xAA6600,		// brown
	0x808080,
	0x808080,
	0x808080,
	0x808080,
};

static char Circle[] = {
	0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x40, 0x00, 0x40, 0x00, 0xF3, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xFC, 0xFC, 0xFC, 0x0C, 0x0C, 0x0C, 0x30, 0x30, 0x30, 0x18, 0x18, 0x18, 0xF0, 0xF0, 0xF0, 0x28,
	0x28, 0x28, 0xB8, 0xB8, 0xB8, 0xE4, 0xE4, 0xE4, 0xD4, 0xD4, 0xD4, 0x98, 0x98, 0x98, 0xC8, 0xC8,
	0xC8, 0x00, 0x00, 0x30, 0xAC, 0xAC, 0xAC, 0x68, 0x68, 0x68, 0x84, 0x84, 0x84, 0x2C, 0x00, 0x00,
	0x00, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 0x04, 0xFE, 0x90, 0x89, 0x49, 0xAB, 0xBD, 0x38, 0xEB,
	0x8D, 0x25, 0xF8, 0x60, 0x28, 0x8E, 0x64, 0x69, 0x9E, 0xE4, 0x84, 0xAE, 0x6C, 0xBB, 0xAA, 0x6E,
	0x2C, 0xB3, 0xF0, 0x6C, 0xDF, 0x60, 0x8D, 0xEF, 0xB1, 0xCE, 0xFF, 0x28, 0x1F, 0x70, 0x38, 0x12,
	0x12, 0x8F, 0x00, 0x23, 0x72, 0xA8, 0x5C, 0xFE, 0x9A, 0xCE, 0x1D, 0x34, 0x7A, 0x9B, 0x52, 0x67,
	0xD6, 0x6B, 0x4F, 0x80, 0x24, 0x18, 0x0C, 0x04, 0x64, 0x76, 0x35, 0x78, 0x2C, 0x12, 0x85, 0x40,
	0xA0, 0x90, 0x58, 0x3C, 0x06, 0xC0, 0xB1, 0x69, 0x70, 0x50, 0xDB, 0xEF, 0xF6, 0x03, 0x5C, 0xCA,
	0xC5, 0x29, 0xD2, 0x78, 0x81, 0x6A, 0x05, 0x0A, 0x7C, 0x37, 0x06, 0x75, 0x82, 0x8A, 0x6A, 0x07,
	0x06, 0x55, 0x7D, 0x32, 0x04, 0x08, 0x8B, 0x94, 0x01, 0x08, 0x61, 0x58, 0x90, 0x2E, 0x92, 0x95,
	0x95, 0x97, 0x99, 0x33, 0x0D, 0x9D, 0x9D, 0x0D, 0xA0, 0x31, 0x0A, 0xA3, 0xA3, 0x85, 0x5B, 0x31,
	0x03, 0x80, 0xA9, 0x94, 0x05, 0x7B, 0x2D, 0x72, 0x00, 0x89, 0xB0, 0x95, 0x07, 0xAC, 0x2D, 0x03,
	0xB8, 0xA9, 0xB3, 0x2F, 0x9A, 0x28, 0x0E, 0xBE, 0xA3, 0x0E, 0x2E, 0x72, 0x0B, 0xC5, 0x9D, 0x0B,
	0xC8, 0xC2, 0x27, 0x09, 0xCB, 0x95, 0x09, 0xCE, 0x2D, 0x04, 0xAF, 0xD2, 0x82, 0x05, 0x98, 0xC1,
	0x2D, 0x06, 0xD9, 0x95, 0x8E, 0x34, 0xCF, 0x25, 0xDF, 0xE0, 0x8B, 0xE2, 0xDD, 0x2C, 0xD7, 0xE7,
	0xDA, 0xDC, 0x41, 0xE4, 0x25, 0xD1, 0xED, 0x78, 0xD4, 0xB4, 0xF1, 0x24, 0xCA, 0xF4, 0x77, 0xCD,
	0xF7, 0x2E, 0x0F, 0xFB, 0xEE, 0x3C, 0xA8, 0xC6, 0x2B, 0xA0, 0x1D, 0x60, 0xF0, 0x62, 0xDC, 0xA2,
	0xA7, 0x8B, 0x20, 0x2F, 0x6C, 0xE0, 0x64, 0xED, 0x72, 0x81, 0x8A, 0xDE, 0x2A, 0x87, 0x2E, 0x16,
	0x4A, 0x6B, 0x38, 0xD1, 0x85, 0x81, 0x49, 0x5E, 0xD9, 0x10, 0xA4, 0xC3, 0xB8, 0x09, 0x64, 0xB1,
	0x4F, 0xA6, 0x66, 0x10, 0x10, 0xE5, 0xAB, 0xC1, 0xBB, 0x8E, 0x33, 0xFE, 0xA4, 0x22, 0x64, 0x68,
	0x07, 0x9D, 0x4E, 0x7A, 0x78, 0xD4, 0x9A, 0xE3, 0xE0, 0x0C, 0x48, 0x36, 0x0B, 0x1C, 0x20, 0x7C,
	0x24, 0xE6, 0xCB, 0xCB, 0x27, 0xF8, 0xB4, 0xA4, 0x54, 0xAA, 0x33, 0x29, 0x53, 0x92, 0x4F, 0x6D,
	0xEC, 0x8C, 0x5A, 0xC4, 0x29, 0xD5, 0x84, 0x57, 0x89, 0x66, 0x95, 0x6A, 0x75, 0x6B, 0x8A, 0xAE,
	0x5E, 0x45, 0x4C, 0xF5, 0x3A, 0x76, 0x6B, 0xD9, 0xAC, 0x67, 0xAF, 0x0A, 0x90, 0xC0, 0xA1, 0xAD,
	0xDB, 0xB7, 0x1C, 0x18, 0x44, 0x00, 0x00, 0x3B, };

static char Square[] = {
	0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x40, 0x00, 0x40, 0x00, 0xF2, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xFC, 0xFC, 0xFC, 0x0C, 0x0C, 0x0C, 0xD4, 0xD4, 0xD4, 0x00, 0x00, 0x30, 0xB8, 0xB8, 0xB8, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 0x03,
	0xFE, 0x48, 0xB2, 0xDC, 0xFE, 0x30, 0xCA, 0x09, 0x15, 0xB8, 0x38, 0xEB, 0xCD, 0xBB, 0xFF, 0xDC,
	0x02, 0x8E, 0x64, 0x39, 0x8A, 0x66, 0xAA, 0x92, 0xE8, 0xEA, 0xBE, 0x58, 0x0B, 0xCF, 0xA9, 0x4C,
	0xDF, 0xA0, 0x8D, 0xEF, 0x9B, 0xCE, 0xFF, 0x00, 0x1F, 0x70, 0x27, 0x1C, 0xDE, 0x8A, 0xC6, 0x19,
	0x32, 0xF9, 0x5A, 0x32, 0x57, 0xCE, 0x67, 0x4D, 0x20, 0xE5, 0x45, 0x2F, 0x85, 0x81, 0x76, 0xCB,
	0xED, 0x7A, 0xBF, 0xDA, 0x82, 0xE9, 0x0A, 0x18, 0x04, 0xCE, 0xE8, 0xB4, 0x7A, 0xCD, 0x3E, 0x0F,
	0xC6, 0xD4, 0x92, 0xB9, 0x4D, 0xAF, 0xBB, 0xE1, 0xA6, 0xB9, 0x7D, 0xAF, 0x7E, 0x97, 0xC8, 0x7A,
	0x7C, 0x82, 0x7E, 0x2C, 0x71, 0x24, 0x81, 0x82, 0x7B, 0x84, 0x27, 0x86, 0x23, 0x88, 0x89, 0x75,
	0x8B, 0x39, 0x8D, 0x20, 0x8F, 0x90, 0x6D, 0x92, 0x1F, 0x80, 0x97, 0x7C, 0x99, 0x1E, 0x9B, 0x9C,
	0x76, 0x9E, 0x1D, 0xA0, 0xA1, 0x74, 0xA3, 0x21, 0x94, 0x1F, 0x96, 0xA6, 0x68, 0xA8, 0x3D, 0xAA,
	0x1E, 0xAC, 0xAD, 0x01, 0xAF, 0x1A, 0xA5, 0xB4, 0x7D, 0x78, 0x72, 0xB9, 0x6C, 0xB6, 0x19, 0xB8,
	0xBD, 0x77, 0x7F, 0xB1, 0x1D, 0xB3, 0xAD, 0xBF, 0x31, 0xC5, 0x1C, 0xC7, 0xA6, 0xC9, 0x17, 0xC1,
	0xC2, 0xCF, 0x41, 0xCB, 0x1B, 0xCD, 0xA1, 0xD3, 0xD1, 0xBD, 0xD9, 0xD5, 0x1A, 0xD7, 0x9C, 0xDC,
	0x79, 0xC2, 0xBA, 0xC4, 0xE2, 0xE3, 0xAE, 0xBB, 0x87, 0xE7, 0xE8, 0xE5, 0xBC, 0xEB, 0xB5, 0xE9,
	0x8E, 0xEF, 0xF0, 0xED, 0xEA, 0xEF, 0xE1, 0xEE, 0xEB, 0xF8, 0xF6, 0xFA, 0xF1, 0x95, 0xF3, 0xFB,
	0x46, 0x64, 0x01, 0x43, 0xB0, 0x60, 0x18, 0x7F, 0x55, 0x60, 0x90, 0x49, 0xA8, 0xA9, 0x1B, 0xC3,
	0x42, 0x0F, 0x69, 0x2C, 0x8C, 0x08, 0x8B, 0xA2, 0x42, 0x87, 0x16, 0x49, 0x61, 0xCC, 0x58, 0x1B,
	0x91, 0xA3, 0x8A, 0x89, 0x1E, 0x41, 0x72, 0x14, 0x99, 0x91, 0xA4, 0x45, 0x93, 0x14, 0x51, 0x46,
	0x14, 0xA0, 0x80, 0x82, 0xCB, 0x97, 0x30, 0x29, 0x10, 0x48, 0x00, 0x00, 0x3B, };

static char Star[] = {
	0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x40, 0x00, 0x40, 0x00, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xFC, 0xFC, 0xFC, 0x0C, 0x0C, 0x0C, 0xF0, 0xF0, 0xF0, 0x34, 0x34, 0x34, 0xB8, 0xB8, 0xB8, 0x30,
	0x30, 0x30, 0xAC, 0xAC, 0xAC, 0x64, 0x64, 0x64, 0x84, 0x84, 0x84, 0x50, 0x50, 0x50, 0xA0, 0xA0,
	0xA0, 0x78, 0x78, 0x78, 0xE4, 0xE4, 0xE4, 0x00, 0x00, 0x30, 0xD4, 0xD4, 0xD4, 0x18, 0x18, 0x18,
	0x90, 0x90, 0x90, 0x28, 0x28, 0x28, 0x5C, 0x5C, 0x5C, 0x68, 0x68, 0x68, 0xCC, 0xCC, 0xCC, 0xC8,
	0xC8, 0xC8, 0x98, 0x98, 0x98, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2C, 0x00, 0x00,
	0x00, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 0x05, 0xFE, 0xA0, 0x23, 0x8C, 0x64, 0x69, 0x9E, 0x68,
	0xAA, 0xAE, 0xA8, 0x08, 0xBC, 0x70, 0x2C, 0xCF, 0x74, 0x6D, 0xDF, 0xF4, 0x88, 0xEF, 0x7C, 0xBF,
	0xEB, 0xBE, 0xA0, 0x90, 0x07, 0x1C, 0x1A, 0x8F, 0xB0, 0x22, 0x72, 0x19, 0x54, 0x32, 0x9F, 0x38,
	0x27, 0x74, 0x3A, 0x93, 0x52, 0xAF, 0x00, 0x2B, 0x76, 0xAA, 0xDD, 0x3E, 0xBB, 0xDE, 0x25, 0x38,
	0x7C, 0x1C, 0x1B, 0x05, 0x86, 0xAD, 0x79, 0x38, 0x49, 0xA8, 0x05, 0xDB, 0x43, 0x03, 0x7E, 0x5D,
	0x07, 0x21, 0x81, 0x80, 0x02, 0x6B, 0xF7, 0x31, 0xF2, 0x0B, 0x7C, 0x74, 0x54, 0x0F, 0x79, 0x03,
	0x82, 0x57, 0x06, 0x79, 0x79, 0x08, 0x75, 0x83, 0x50, 0x17, 0x8B, 0x01, 0x05, 0x8E, 0x54, 0x02,
	0x0D, 0x92, 0x03, 0x10, 0x96, 0x8F, 0x4C, 0x0A, 0x92, 0x8C, 0x9C, 0x54, 0x05, 0xA0, 0x01, 0x15,
	0xA2, 0x50, 0x78, 0xA5, 0x01, 0x12, 0x5C, 0x9D, 0x42, 0x02, 0x04, 0xB2, 0x11, 0xAB, 0x01, 0x11,
	0xB2, 0x04, 0xAF, 0x4D, 0xBA, 0x33, 0x0A, 0x07, 0x07, 0x03, 0xC1, 0xC2, 0x03, 0xB5, 0xB5, 0xC3,
	0xC2, 0xBF, 0x7B, 0x3F, 0xBC, 0x33, 0x14, 0xC4, 0xC5, 0xD0, 0xD0, 0x03, 0x14, 0x3D, 0x7D, 0x06,
	0x85, 0xD1, 0xD9, 0x92, 0x0F, 0x69, 0xD5, 0xCC, 0x39, 0x07, 0xDA, 0xDA, 0x07, 0xDF, 0x55, 0xE5,
	0x34, 0x08, 0xE2, 0xD0, 0x8D, 0xB0, 0xE7, 0x34, 0x04, 0xCF, 0xEA, 0x8B, 0x03, 0x04, 0x67, 0xEE,
	0x39, 0xA4, 0xF2, 0x93, 0xF7, 0xE6, 0x4B, 0x7F, 0xEA, 0x0C, 0xC4, 0xF0, 0xAB, 0xF1, 0x2F, 0x5B,
	0x40, 0x81, 0x4C, 0x16, 0x88, 0x0B, 0x84, 0x10, 0x89, 0x80, 0x78, 0xD1, 0x06, 0x0C, 0xEC, 0x87,
	0x84, 0x80, 0xBC, 0x7A, 0x0E, 0x27, 0xCA, 0x48, 0x20, 0xCF, 0x4D, 0xC6, 0x25, 0x16, 0x56, 0x0D,
	0xC0, 0x06, 0xCA, 0x42, 0xC3, 0x33, 0xAB, 0x79, 0x0A, 0x48, 0x10, 0xC0, 0xB1, 0x94, 0xC6, 0x24,
	0x2F, 0x5F, 0x4C, 0x00, 0x35, 0x80, 0xDD, 0x0B, 0x05, 0x98, 0x24, 0x4D, 0xF8, 0x78, 0x44, 0xE1,
	0xA2, 0x02, 0x9B, 0x64, 0x08, 0xF0, 0x09, 0x88, 0xA7, 0x91, 0x9C, 0x01, 0x6C, 0xD2, 0x50, 0xF0,
	0xAC, 0x81, 0x51, 0x21, 0x8A, 0x4C, 0x75, 0xBB, 0x21, 0x21, 0xDF, 0xD4, 0x76, 0x47, 0x12, 0x0C,
	0x48, 0xC0, 0x0F, 0xC1, 0xD6, 0x32, 0x31, 0x6F, 0x0D, 0x21, 0x10, 0x01, 0x2C, 0x19, 0x44, 0x67,
	0x51, 0xA5, 0x85, 0xD2, 0x67, 0x6D, 0x94, 0x98, 0x6E, 0x85, 0xC2, 0x8D, 0x0B, 0x93, 0x2E, 0x93,
	0xB6, 0x76, 0x29, 0xE6, 0xB5, 0xB7, 0xD7, 0x6C, 0xDF, 0x21, 0x78, 0xFF, 0x66, 0x99, 0x6B, 0x37,
	0xF0, 0x5F, 0xC3, 0x7D, 0x11, 0xEF, 0x15, 0x20, 0x82, 0x85, 0xE3, 0xC7, 0x90, 0x59, 0x38, 0x08,
	0x01, 0x00, 0x3B, };

static char Triangle[] = {
	0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x40, 0x00, 0x40, 0x00, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xFC, 0xFC, 0xFC, 0x0C, 0x0C, 0x0C, 0x5C, 0x5C, 0x5C, 0x18, 0x18, 0x18, 0xF0, 0xF0, 0xF0, 0xD4,
	0xD4, 0xD4, 0xE4, 0xE4, 0xE4, 0x00, 0x00, 0x30, 0xB8, 0xB8, 0xB8, 0x40, 0x40, 0x40, 0xC8, 0xC8,
	0xC8, 0x50, 0x50, 0x50, 0xA0, 0xA0, 0xA0, 0x78, 0x78, 0x78, 0xCC, 0xCC, 0xCC, 0x30, 0x30, 0x30,
	0xAC, 0xAC, 0xAC, 0x34, 0x34, 0x34, 0x98, 0x98, 0x98, 0x90, 0x90, 0x90, 0x84, 0x84, 0x84, 0x28,
	0x28, 0x28, 0x68, 0x68, 0x68, 0x64, 0x64, 0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2C, 0x00, 0x00,
	0x00, 0x00, 0x40, 0x00, 0x40, 0x00, 0x00, 0x05, 0xFE, 0x20, 0x22, 0x8C, 0x64, 0x69, 0x9E, 0x68,
	0xAA, 0xAE, 0xA8, 0x08, 0xBC, 0x70, 0x2C, 0xCF, 0x74, 0x6D, 0xDF, 0xF4, 0x88, 0xEF, 0x7C, 0xBF,
	0xEB, 0xBE, 0xA0, 0x90, 0x07, 0x1C, 0x1A, 0x8F, 0xB0, 0x22, 0x72, 0x19, 0x54, 0x32, 0x9F, 0x38,
	0x27, 0x74, 0x3A, 0x93, 0x52, 0xAF, 0x00, 0x2B, 0x76, 0xAA, 0xDD, 0x3E, 0xBB, 0xDE, 0x25, 0x38,
	0x7C, 0x1C, 0x1B, 0x09, 0x98, 0xAD, 0x79, 0x18, 0x31, 0x08, 0xB0, 0xEB, 0xA0, 0x22, 0x10, 0xA8,
	0xC0, 0xDF, 0xD7, 0x07, 0xBD, 0x40, 0xB8, 0xC6, 0x7B, 0x17, 0x74, 0x74, 0x11, 0x7E, 0x78, 0x50,
	0x02, 0x05, 0x82, 0x74, 0x10, 0x54, 0x7F, 0x3B, 0x0D, 0x8A, 0x74, 0x0F, 0x8D, 0x86, 0x4C, 0x16,
	0x91, 0x82, 0x0C, 0x5C, 0x95, 0x4B, 0x0B, 0x98, 0x74, 0x6E, 0x87, 0x9C, 0x47, 0x03, 0x9F, 0x82,
	0x0E, 0xA2, 0x5F, 0x07, 0xA6, 0x74, 0x07, 0x7D, 0x4C, 0x8E, 0x35, 0x13, 0xAC, 0x82, 0x13, 0x5F,
	0xA3, 0x42, 0x04, 0x89, 0xB4, 0x01, 0x05, 0x8C, 0x62, 0xB8, 0x41, 0x09, 0xBC, 0x82, 0x09, 0xB0,
	0xC1, 0x3D, 0x0C, 0xC4, 0x8A, 0x9A, 0x48, 0xB1, 0x32, 0x06, 0xCB, 0x82, 0x06, 0xC0, 0x48, 0x0E,
	0xD2, 0x8A, 0xA8, 0x65, 0xC8, 0x38, 0xBA, 0xD8, 0x82, 0x7C, 0xDB, 0x47, 0x90, 0xDF, 0x82, 0x0D,
	0xE2, 0x43, 0x12, 0xE5, 0x91, 0x12, 0x46, 0xCF, 0x00, 0x9E, 0xEB, 0x82, 0x0B, 0xEE, 0xDC, 0x35,
	0xA5, 0xF2, 0x8A, 0x03, 0x43, 0xCF, 0xAB, 0xF9, 0x82, 0x07, 0xF8, 0xD9, 0x9B, 0x41, 0xE1, 0x5F,
	0x24, 0x0A, 0x42, 0x1C, 0x79, 0x33, 0x08, 0xEE, 0x55, 0x0F, 0x47, 0xC3, 0x18, 0x2A, 0x32, 0xE6,
	0xE3, 0xCF, 0x1C, 0x89, 0x91, 0x14, 0x54, 0x1C, 0xF8, 0x42, 0x40, 0x34, 0x8C, 0x8A, 0x42, 0x11,
	0xE1, 0x98, 0x65, 0x80, 0xC9, 0x93, 0x28, 0x47, 0x53, 0xAA, 0x5C, 0x89, 0x92, 0xE4, 0x3B, 0x32,
	0x37, 0x5E, 0xC2, 0xAC, 0x21, 0x73, 0x66, 0x15, 0x92, 0x36, 0x7F, 0xE0, 0xCC, 0x19, 0x73, 0x27,
	0x4F, 0x9A, 0x3E, 0x7F, 0xDE, 0x14, 0x5A, 0x8D, 0x28, 0x3A, 0xA3, 0x02, 0x91, 0xD6, 0x53, 0x9A,
	0x94, 0x69, 0x93, 0xA0, 0x4A, 0x6B, 0x0A, 0x95, 0xFA, 0x93, 0x2A, 0x4F, 0xAB, 0x39, 0xB1, 0xDA,
	0xD4, 0x3A, 0x93, 0x2B, 0x4C, 0x01, 0x22, 0x58, 0x88, 0x1D, 0x4B, 0x96, 0x05, 0x82, 0x10, 0x00,
	0x3B, };

static char *ImagePtr = NULL;
static int	ImageSize = 0;
static int NumColours = 0;
static int	RefreshDelay=0;
static int	TableNumR = 6;
static int	TableNumC = 9;

static int StaticPage() {
	int i, v, ind;
	int r, c;
	char	tbuf[16];
	
	printf( "Content-type: text/html\n" );
	printf( "Pragma: no-cache\n" );
	printf( "Expires: 0\n" );
	printf( "\n" );
	printf( "<html>\n" );
	printf( "<head>\n" );
	if (RefreshDelay) printf( "<META HTTP-EQUIV=\"Refresh\" CONTENT=\"%d\">\n", RefreshDelay );
	printf( "</head>\n" );
	printf( "<body bgcolor=\"#000000\" text=#FFFFD0 link=\"#FFFF40\" vlink=\"#FFFFFF\">\n" );
//	printf( "<body>\n" );
	printf( "<center>\n" );

	printf( "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n" );
	i = time( NULL );
//	i=0;
	for (r=0; r<TableNumR; r++) {
		printf( "<tr>" );
		for (c=0; c<TableNumC; c++) {
			ind = i % (TotalMod*100);
			
//			printf( "<td bgcolor=\"#%06X\">", BGColours[ (ind / BGColourDiv) % BGColourMod ] );
			printf( "<td bgcolor=\"#C0C0C0\">" );

			printf( "<img src=\"%s?%d\" height=\"64\" width=\"64\"></td>\n", MyURL, ind );
			i+=1+ColourDiv+BGColourDiv;
//			i++;
		}
		printf( "</tr>" );
//		i -= 5;
	}

	
	printf( "</table>\n" );
	printf( "</center>\n" );
	printf( "</body>\n" );
	printf( "</html>\n" );
	exit( 0 );
}

static int MakeImage( int ind ) {
	int i, q, qb, n;
	ulong_t	c, bg, v, r;
	uchar_t	*p;

	switch (ind % ImageMod) {
		case 0:
			ImagePtr = Circle;
			ImageSize = sizeof( Circle );
			NumColours = 16;
			break;
		case 1:
			ImagePtr = Square;
			ImageSize = sizeof( Square );
			NumColours = 8;
			break;
		case 2:
			ImagePtr = Star;
			ImageSize = sizeof( Star );
			NumColours = 32;
			break;
		case 3:
			ImagePtr = Triangle;
			ImageSize = sizeof( Triangle );
			NumColours = 32;
			break;
	}
	
	p = malloc( ImageSize );
	memcpy( p, ImagePtr, ImageSize );
	ImagePtr = p;
	
	c = Colours[ (ind / ColourDiv) % ColourMod ];
	bg = BGColours[ (ind / BGColourDiv) % BGColourMod ];
	p = ImagePtr + 0x10 - 3;
	for (i=0; i<NumColours; i++) {
		for (n=16; n>=0; n-=8 ) {
			q = (c>>n) & 0xFF;
			qb = (bg>>n) & 0xFF;
//			if (q != 0xFF) {
				v = *p;
				r = ((v * q) + ((0xFF - v) * qb)) >> 8;
				if (r > 0xFF) r = 0xFF;
				*p = r;
//			}
			p++;
		}
	}
	
	printf( "Content-type: image/gif\n" );
	printf( "Content-Length: %d\n", ImageSize );
	printf( "Pragma: no-cache\n" );
	printf( "Expires: 0\n" );
	printf( "\n" );
	fflush( stdout );
	write( 1, ImagePtr, ImageSize );
	fflush( stdout );
	exit( 0 );
}

int cs_main( int argc, char *argv[], char *envp[] ){
	char tbuf[1024];
	time_t	tod;
	FILE *f;
	int i, ind=0, q, qb, n;
	char *s;

//	sprintf( MyURL, "%s", getenv( "SCRIPT_NAME" ) );
	sprintf( MyURL, "/cs" );
	
	for (n=1; n<=argc; n++) {
		s = argv[n];
		switch (*s) {
			case 'd':
				RefreshDelay = strtoul( s+1, NULL, 0 );
				break;
			case 'r':
				TableNumR = strtoul( s+1, NULL, 0 );
				break;
			case 'c':
				TableNumC = strtoul( s+1, NULL, 0 );
				break;
			case 'S':	
				TableNumR = 4;
				TableNumC = 5;
				break;
			case 'B':	
				TableNumR = 15;
				TableNumC = 23;
				break;
			default:
				if (isdigit( *s )) {
					ind = strtoul( s, NULL, 0 );
					MakeImage( ind );
				}
		}
	}

	StaticPage();


	exit( 0 );
} 



























