/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"
#define	WODDBUF_BASEOFFSET	1024

int ClearBuffer( Buffer_t *bptr ) {
	if (bptr->ptr == NULL) _BufferResize( bptr, 0 );
	bptr->base = 0;
	bptr->read = bptr->write = bptr->verify = bptr->base + WODDBUF_BASEOFFSET;
	if (bptr->ptr) *(_BufWritePtr( bptr )) = 0;
	return 0;
}

int FreeBuffer( Buffer_t *bptr ) {
	if (bptr->ptr == NULL) return 0;
	free( bptr->ptr );
	bptr->ptr = NULL;
	bptr->base = bptr->max = bptr->flags = 0;
	bptr->read = bptr->write = bptr->verify = bptr->base;
	return 0;
}

int _BufferResize( Buffer_t *bptr, int m ) {
	int v;
	
	v = (m + 1024 + 255);
	for (m = 4096;m<v; m<<=1);
	if (m < bptr->max) return -1;
	if (bptr->max == 0) {
		bptr->ptr = malloc( m );
		ClearBuffer( bptr );
	} else {
		bptr->ptr = realloc( bptr->ptr, m );
	}
	
// actual size may be larger 
	bptr->max = malloc_size( bptr->ptr );
	
//	bptr->max = m;
	return m;
}

int _BufferEnsureSpace( Buffer_t *bptr, int n ) {
	int m, v, t, bhead;
	
	m = (bptr->write - bptr->base) + n;
	if (m < bptr->max) return 0;
	
	if (bptr->flags & BufMustVerify) bhead = bptr->verify;
	else bhead = bptr->read;
	
	t = bptr->write - bhead;

	if (((t+n+WODDBUF_BASEOFFSET) < (bptr->max)) &&	
		((bhead - bptr->base) > ((bptr->max / 2) + WODDBUF_BASEOFFSET))) {
		uchar_t *s, *d;
		
		t = bptr->write - bhead;
		s = _BufReadPtr( bptr );
		v = (bhead - bptr->base) - WODDBUF_BASEOFFSET;
		bptr->base += v;
		d = _BufReadPtr( bptr );
		memcpy( d, s, t+1 );
		return m;
	} 
	return _BufferResize( bptr, m );
}

int BufferEnqueue( Buffer_t *bptr, uchar_t *p, int n ) {
	if (!bptr->max) _BufferResize( bptr, 0 );
	if (n > _BufFreeSpace( bptr )) _BufferEnsureSpace( bptr, n );
	if (p) memcpy( _BufWritePtr( bptr ), p, n );
	bptr->write += n;
	*(_BufWritePtr(bptr)) = 0;
	return n;
}

char *BufferMakeSpace( Buffer_t *bptr, int n ) {
	
	if (n > _BufFreeSpace( bptr )) _BufferEnsureSpace( bptr, n );
	return _BufWritePtr( bptr );
}

int BufferFreeSpace( Buffer_t *bptr ) {
	if (bptr->ptr == NULL) return 0;
	return _BufFreeSpace( bptr );
}

int BufferLength( Buffer_t *bptr ) {
	if (bptr->ptr == NULL) return 0;
	return _BufSize( bptr );
}

char *BufferDequeue( Buffer_t *bptr, int n ) {
	char *p;
	int m;
	
	if (n > _BufSize( bptr )) n = _BufSize( bptr );
	p = _BufReadPtr( bptr );
	bptr->read += n;
	if ((bptr->flags & BufMustVerify) == 0) bptr->verify += n;
	return p;
}

int BufferInject( Buffer_t *bptr, uchar_t *p, int n ) {
	int m;
	
	if ((bptr->read - bptr->base) < n) {
		BufferEnqueue( bptr, NULL, n );
		memmove( _BufReadPtr( bptr ) + n, _BufReadPtr( bptr ), n );
	} else {
		bptr->read -= n;
	}
	if (p) memcpy( _BufReadPtr( bptr ), p, n );
	return n;
}

int BufferDelete( Buffer_t *bptr, int n ) {
	BufferDequeue( bptr, n );
	return n;
}

int BufferMove( Buffer_t *dbptr, Buffer_t *sbptr ) {
	Buffer_t	tb;
	
	memcpy( &tb, sbptr, sizeof( Buffer_t ) );
	memcpy( sbptr, dbptr, sizeof( Buffer_t ) );
	memcpy( dbptr, &tb, sizeof( Buffer_t ) );
	ClearBuffer( sbptr );
	return( BufferLength( dbptr ) );	
}

int bprintf( Buffer_t *bptr, const char *frmt, ... ) {
	va_list	arglist;
	int l;
	char tbuf[1024];
	
	va_start( arglist, frmt );
	
	l = vsnprintf( tbuf, sizeof( tbuf ), frmt, arglist );
	return BufferEnqueue( bptr, tbuf, l );
}

int biprintf( Buffer_t *bptr, const char *frmt, ... ) {
	va_list	arglist;
	int l;
	char tbuf[1024];
	
	va_start( arglist, frmt );
	
	l = vsnprintf( tbuf, sizeof( tbuf ), frmt, arglist );
	return BufferInject( bptr, tbuf, l );
}

