/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"

char *ChTypeStr( int t ) {
	switch (t) {
		case ChType_Listen:	return "listen";
		case ChType_KICK:	return "KICK";
		case ChType_TCP:	return "TCP";
		case ChType_PIPE:	return "PIPE";
	}
	return "unknown";
}

char ChTypeChar( int t ) {
	switch (t) {
		case ChType_Listen:	return 'L';
		case ChType_KICK:	return 'K';
		case ChType_TCP:	return 'T';
		case ChType_PIPE:	return 'P';
	}
	return '?';
}

int ChannelAttachFD( Channel_t *ChPtr, int fd ) {
	int i;
	
	for (i=0; i<MAX_other_fds; i++) {
		if (ChPtr->other_fds[i]) continue;
		ChPtr->other_fds[i] = fd;
		return 0;
	}
	return -1;
}

int ChannelRemoveFD( Channel_t *ChPtr, int fd ) {
	int i;
	
	for (i=0; i<MAX_other_fds; i++) {
		if (ChPtr->other_fds[i]!=fd) continue;
		ChPtr->other_fds[i] = 0;
		return 0;
	}
	return -1;
}

int ChannelAttachPTR( Channel_t *ChPtr, void *ptr ) {
	int i;
	
	for (i=0; i<MAX_other_ptr; i++) {
		if (ChPtr->other_ptr[i]) continue;
		ChPtr->other_ptr[i] = ptr;
		return 0;
	}
	return -1;
}

int ChannelRemovePTR( Channel_t *ChPtr, void *ptr ) {
	int i;
	
	for (i=0; i<MAX_other_ptr; i++) {
		if (ChPtr->other_ptr[i]!=ptr) continue;
		ChPtr->other_ptr[i] = NULL;
		return 0;
	}
	return -1;
}

int ChannelDistroy( Channel_t *ChPtr, int forced ) {
	int	t, i;
	Channel_t	*PChPtr;
	if ((verbose&4) || forced) printf( "ChDis#%d %s\n", ChPtr->idx, ChTypeStr( ChPtr->type ) );

	kill_thread_zombie( ChPtr );
	if (ChPtr->thread_pid) return 0;

	if (forced) forced=2;

	if (ChPtr->create_time) {
		ChPtr->total_time = SetCurrentTime() - ChPtr->create_time;
		if (ChPtr->total_time == 0) ChPtr->total_time++;
	}
	
	if (ChannelEvent( ChPtr, CHEV_Close, forced )) return 0;
	

	if (ChPtr->SPtr) {
		ChPtr->SPtr->ChildPtr = NULL;
		ChPtr->SPtr = NULL;
	}
	if (ChPtr->ChildPtr) {
		ChPtr->ChildPtr->SPtr = NULL;
		ChPtr->ChildPtr = NULL;
	}
	if (ChPtr->fdr) {
		RFDclose( ChPtr->fdr );
		if (ChPtr->fdw == ChPtr->fdr) ChPtr->fdw=0;
		ChPtr->fdr = 0;
	}
	if (ChPtr->fdw) {
		RFDclose( ChPtr->fdw );
		ChPtr->fdw = 0;
	}
	if (ChPtr->fde) {
		RFDclose( ChPtr->fde );
		ChPtr->fde = 0;
	}
	for (i=0; i<MAX_other_fds; i++) {
		if (!ChPtr->other_fds[i]) continue;
		RFDclose( ChPtr->other_fds[i] );
		ChPtr->other_fds[i] = 0;
	}
	for (i=0; i<MAX_other_ptr; i++) {
		if (ChPtr->other_ptr[i] == NULL) continue;
		free( ChPtr->other_ptr[i] );
		ChPtr->other_ptr[i] = NULL;
	}
	
	SetRF( ChPtr, 0, -1 );
#if _Woddity_ChannelD_FreeBuffer
	FreeBuffer( &(ChPtr->o) );
	FreeBuffer( &(ChPtr->i) );
	FreeBuffer( &(ChPtr->e) );
#else
#if _Woddity_ChannelD_ClearBuffer
	ClearBuffer( &(ChPtr->o) );
	ClearBuffer( &(ChPtr->i) );
	ClearBuffer( &(ChPtr->e) );
#endif
#endif

	if (ChPtr->SPtr) {
		if (ChPtr->SPtr->ChildPtr == ChPtr) ChPtr->SPtr->ChildPtr = NULL;
		ChPtr->SPtr = NULL;
	}

	ChPtr->type = 0;
	return 0;
}

int ChannelEvent( Channel_t *ChPtr, unsigned chev, int stat ) {
	ChPtr->access_time = GetCurrentTime();
	
	if (verbose&8) printf( "ChannelEvent #%d %d\n", ChPtr->idx, chev );

	if (ChPtr->chev_fp) return (ChPtr->chev_fp)(ChPtr, chev, stat );
	return -1;
}

static int NeedChannelHousekeeping;
static int LastChannelHousekeeping;

int ChannelHousekeeping() {
	Channel_t	*ChPtr;
	int i, tm, tv;
	
	tm = GetCurrentTime();

	NeedChannelHousekeeping = 1000;
	LastChannelHousekeeping = tm;
	
	for (i=NumChannels, ChPtr = Channels + i; i>0; i--, ChPtr--) {
		if (ChPtr->type) continue;
		if (ChPtr->i.ptr == NULL) continue;
		if ((tv = tm - ChPtr->access_time) < CHANNEL_ATROPHY) continue;
		FreeBuffer( &(ChPtr->o) );
		FreeBuffer( &(ChPtr->i) );
		FreeBuffer( &(ChPtr->e) );
	}
	return 0;
	
}


static int __LastChannelCreate=0;

Channel_t *ChannelCreate( Channel_t *PrChPtr, int type, ChEventCallback_t  *chev_fp ) {
	Channel_t	*ChPtr=NULL;
	int i;
	
	if (--NeedChannelHousekeeping <= 0) ChannelHousekeeping();
	if ((LastChannelHousekeeping == 0) ||
		((GetCurrentTime() - LastChannelHousekeeping) > (WTSecond * 7))) ChannelHousekeeping();

#if 0
	for (ChPtr = Channels+__LastChannelCreate, i=__LastChannelCreate; i<MaxChannels; i++, ChPtr++) {
		if (i >= NumChannels) {
			ChPtr = NULL;
			break;
		}
		if (ChPtr->type) continue;
		break;
	}
#endif

//	if (ChPtr == NULL) {	
		for (ChPtr = Channels, i=0; i<MaxChannels; i++, ChPtr++) {
			if (i >= MaxChannels) return NULL;
			if (ChPtr->type) continue;
			break;
		}
//	}
	
	ChPtr->type = type;
	ChPtr->idx=i+1;
	__LastChannelCreate=i+1;

	if (i >= NumChannels) {
		NumChannels = i+1;
		if (NumChannels > MaxNumChannels) MaxNumChannels = NumChannels;
	}

	if (verbose&4) printf( "ChannelCreate #%d %s\n", i+1, ChTypeStr( ChPtr->type ) );

	ChPtr->dname[0]=0;
	ChPtr->content_offset = ChPtr->content_length = ChPtr->serve_type = 0;
	
	ChPtr->addr_len = 0;
	memset( &(ChPtr->addr), 0, sizeof(ChPtr->addr) );
	ChPtr->req_status = 0;
	ChPtr->req_log[0] = 0;
	ChPtr->total_write = 0;
	ChPtr->wservptr = NULL;
	ChPtr->VTPtr = NULL;
	ChPtr->VSPtr = NULL;
	ChPtr->cxcb_fp = NULL;
	ChPtr->chev_fp = NULL;
	ChPtr->expire = 0;
	ChPtr->root = NULL;
	ChPtr->enc = ChPtr->encl = 0;
	ChPtr->argc = ChPtr->argcl = 0;

	SetRF( ChPtr, RF_Read, 0 );
	ChPtr->create_time = ChPtr->access_time = GetCurrentTime();
	ChPtr->expire = 0;
	ClearBuffer( &(ChPtr->o) );
	ClearBuffer( &(ChPtr->i) );
	ClearBuffer( &(ChPtr->e) );

	if (PrChPtr) {
		ChPtr->VSPtr = PrChPtr->VSPtr;
	}
	ChPtr->chev_fp = chev_fp;
	ChannelEvent( ChPtr, CHEV_Init, 0 );
	return ChPtr;
}

Channel_t *ChannelIndex( int idx ) {
	Channel_t	*ChPtr;
	int i;
	
	if ((idx <= 0) || (idx > MaxChannels)) return NULL;
	return (Channels + idx-1);
}


