/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"

static char *WWWRoot="/var/www";
static char *CGIRoot="/usr/lib/cgi-bin";
static char *USERRoot="/home/$/.html";
// static char *LOGRoot="/var/log/woddity";
static char *ServerAdmin;
static char *ServerName;
static char *DefaultPage="index.html";
static int ChildExpire=2*60;



typedef struct {
	char	*id;
	int	flags;
	void	*fp;
} HTTPHandler_t;

#define	HH_MASK		0xFF

#define	MAX_HTTPHandlers	256

HTTPHandler_t HTTPHandlers[MAX_HTTPHandlers] = {
	"file",		HHService | WS_Thread,	(void *)HTTPReadFile,
	"exec",		HHService | WS_Inline,	(void *)HTTPExec,
	"alias",	HHService | WS_Inline,	(void *)HTTPAlias,
	"error",	HHService | WS_Inline,	(void *)HTTPError,
	"robots",	HHService | WS_Inline,	(void *)HTTPRobots,
	"status",	HHService | WS_Inline,	(void *)MakeStatusPage,
	"policy",	HHPolicy,	NULL,
	"localonly",	HHLocalOnly,	NULL,
	NULL, 0, NULL,
};

int ConfigWoddityAppend( char *id, unsigned flags, void *fp ) {
	HTTPHandler_t	*HPtr;
	int i;
	
	for (i=0, HPtr = HTTPHandlers; i<MAX_HTTPHandlers; i++, HPtr++) {
		if (HPtr->id) continue;
		HPtr->id = strperm( id );
		HPtr->flags = flags;
		HPtr->fp = fp;
		return i;
	}
	fprintf( stderr, "*** Too Many ConfigWoddityAppend's, num %d ***\n", i );
	return -1;
}

static int LConf_Base_RFlags=0;

static VServer_t	*CurConfVSPtr=NULL;

VServer_t	*NewVServer() {
	VServer_t	*VSPtr;
	int i;
	
	for (i=0; ; i++) {
		if (i >= MAX_VSERVER) {
			fprintf( stderr, "*** TOO MANY VIRTUAL SERVERS (max: %d) ***\n", MAX_VSERVER );
			exit( 1 );
		}
		if (VServers[i] == NULL) break;
	}
	VSPtr = VServers[i] = calloc( 1, sizeof( VServer_t ) );
	CurConfVSPtr = VSPtr;
	malloc_lock( VSPtr, 1 );
	VSPtr->flags = VS_Valid;
	LConf_Base_RFlags=0;
	return VSPtr;
}

int LConfWoddity( VServer_t *VSPtr, char *s ) {
	HTTPHandler_t	*HHPtr;
	WService_t	*wp = NULL;
	char	*t, *q;
	char	*rtype, *rmatch, *rextra, *ralias, *rroot;
	int	rflags;
	char	rsave[256];

	if (VSPtr) CurConfVSPtr = VSPtr;
	else VSPtr = CurConfVSPtr;
		
	if (t = strchr( s, '#')) *t = 0;
	while (*s && isspace( *s )) s++;
	if (!*s) return 0;
	
	if (verbose & 4) printf( "LConfWoddity: %s\n", s );
	
	if (*s == '!') {		// new vserver
		rtype = ++s;
		while (*s && !isspace( *s )) s++;
		*s++ = 0;
		while (*s && isspace( *s )) s++;
		q = s;
		while (*s && !isspace( *s )) s++;
		*s++ = 0;
		if (!stricmp( rtype, "http" )) {
			VSPtr = NewVServer();
			VSPtr->vsinit_fp = AttachHTTP_Listen;
			VSPtr->Port = strperm( q );
			VSPtr->ServerName = ServerName;
			VSPtr->ServerAdmin = ServerAdmin;
			VSPtr->WWWRoot = WWWRoot;
			VSPtr->CGIRoot = CGIRoot;
			VSPtr->USERRoot = USERRoot;
			VSPtr->LOGRoot = LOGRoot;
			VSPtr->LOGName = NULL;
			VSPtr->DefaultPage = DefaultPage;
			VSPtr->ChildExpire = ChildExpire;
		}
	} else if (t = strchr( s, '=')) {	// assignment
		q = t-1;
		*t++ = 0;
		while (*t && isspace( *t )) t++;	
		while ((q > s) && isspace( *q )) *q-- = 0;
		q = t + strlen( t ) -1;
		while ((q > t) && isspace( *q )) *q-- = 0;
		
		if (verbose & 2) printf( "assign %s = %s\n", s, t );

		if (!stricmp( s, "MaxChildren")) {
			MaxChildren = strtol( t, NULL, 0 );

		} else if (!stricmp( s, "verbose")) {
			verbose = strtol( t, NULL, 0 );

		} else if (!stricmp( s, "ServerName")) {
			if (VSPtr) VSPtr->ServerName = strperm( t );
			else ServerName = strperm( t );

		} else if (!stricmp( s, "ServerAdmin")) {
			if (VSPtr) VSPtr->ServerAdmin = strperm( t );
			else ServerAdmin = strperm( t );

		} else if (!stricmp( s, "WWWRoot")) {
			if (VSPtr) VSPtr->WWWRoot = strperm( t );
			else WWWRoot = strperm( t );

		} else if (!stricmp( s, "CGIRoot")) {
			if (VSPtr) VSPtr->CGIRoot = strperm( t );
			else CGIRoot = strperm( t );

		} else if (!stricmp( s, "USERRoot")) {
			if (VSPtr) VSPtr->USERRoot = strperm( t );
			else USERRoot = strperm( t );

		} else if (!stricmp( s, "LOGRoot")) {
			if (VSPtr) VSPtr->LOGRoot = strperm( t );
			else LOGRoot = strperm( t );

		} else if (!stricmp( s, "LOGName")) {
			if (VSPtr) VSPtr->LOGName = strperm( t );
			else fprintf( stderr, "*** LOGName before Server ignored ***" );

		} else if (!stricmp( s, "DefaultPage")) {
			if (VSPtr) VSPtr->DefaultPage = strperm( t );
			else DefaultPage = strperm( t );

		} else if (!stricmp( s, "ChildExpire")) {
			if (VSPtr) VSPtr->ChildExpire = strtol( t, NULL, 0 );
			else ChildExpire = strtol( t, NULL, 0 );
		}
	} else {
						// rules
//		rsave = strdup( s );
		strncpy( rsave, s, sizeof( rsave ) );
		
		rtype = rmatch = rextra = ralias = rroot = NULL;
		rflags = LConf_Base_RFlags;
	
		rtype = s;
		while (s && *s) {
			while (*s && !isspace( *s )) s++;
			*s++ = 0;
			while (*s && isspace( *s )) s++;
			if (*s == '-') {
				for (s++; *s && !isspace( *s );) {
					switch (*s++) {
						case 'Q':	rflags |= WS_CGIQuery;	continue;
						case 'T':	rflags |= WS_Thread;	rflags &= ~WS_Inline;	continue;
						case 'I':	rflags |= WS_Inline;	rflags &= ~WS_Thread;	continue;
						case 'A':	rflags &= ~(WS_Thread | WS_Inline);	continue;
						case 'C':	rflags |= WS_CGIRoot;	rflags &= ~WS_USERRoot;	continue;
						case 'W':	rflags &= ~(WS_CGIRoot | WS_USERRoot);	continue;
						case 'U':	rflags &= ~WS_CGIRoot;	rflags |= WS_USERRoot;	continue;
						case 'a':				
							while (*s && isspace( *s )) s++;
							ralias = s;
							break;
						case 'r':				
							while (*s && isspace( *s )) s++;
							rroot = s;
							break;
					}
					break;
				}
			} else {
				if (!rmatch) rmatch = s;
				else if (!rextra) rextra = s;
			}
		}
		if (verbose & 2) printf( "rule %s %s %s %X\n", rtype, rmatch, rextra, rflags );
		
		for (HHPtr = HTTPHandlers; ; HHPtr++) {
			if (HHPtr->id == NULL) {
				fprintf( stderr, "*** UNKNOWN HTTP Service %s ***\n", rtype );
				break;
			}
			if (strcmp( HHPtr->id, rtype )) continue;
			switch (HHPtr->flags & HH_MASK) {
				case HHPolicy:
					LConf_Base_RFlags = rflags;
					break;
				case HHLocalOnly:
					if (VSPtr == NULL) {
						fprintf( stderr, "*** LocalOnly flag before Server ***\n" );
						break;
					}
					VSPtr->flags |= VS_LOCALONLY;
					break;
				case HHService:
					if (VSPtr == NULL) {
						fprintf( stderr, "*** Rule before Server %s ***\n", rtype );
						break;
					}
					wp = AttachHTTPService( VSPtr, HHPtr->fp, rmatch, rextra, rflags );
					break;
				case HHBuiltin:
					if (VSPtr == NULL) {
						fprintf( stderr, "*** Rule before Server %s ***\n", rtype );
						break;
					}
					wp = AttachHTTPBuiltin( VSPtr, HHPtr->fp, rmatch, rextra, rflags );
					break;
			}	
			break;
		}
		
		if (wp && ((wp->flags & (WS_Thread | WS_Inline)) == 0)) wp->flags |= HHPtr->flags & WS_Thread | WS_Inline;
		if (wp && ralias && *ralias) wp->alias = strperm( ralias );
		if (wp && rroot && *rroot) wp->root = strperm( rroot );

		if (wp) {
//			if (wp->rule) free( wp->rule );
			wp->rule = strperm( rsave );
//			rsave = NULL;
		}
	}	
//	if (rsave) free( rsave );
	return 0;
}

int ConfigWoddity( char *flnm ) {
	char	tbuf[1024];
	FILE	*fp;
	
	if ((fp = fopen( flnm, "r" )) == NULL) {
		perror( "Unable to open config file" );
		return -1;
	}
	while (fgets( tbuf, sizeof( tbuf ), fp )) {
		LConfWoddity( NULL, tbuf );
	}
	fclose( fp );
}

static char *_StandardConfigWoddity_Input[] = {
	"ServerName = robotest.woddity",
	"WWWRoot = /var/www",
	"CGIRoot = /usr/lib/cgi-bin",
	"LOGRoot = /var/log/woddity",
	"DefaultPage = status",

	"file	*.html	text/html",
	"file	*.htm	text/html",
	"file	*.gif	image/gif",
	"file	*.png	image/png",
	"file	*.jpg	image/jpg",
	"file	*	text/plain",
	
	"exec	cgi-bin/*	-C",

	"status	status	",
	NULL	};
	
int StandardConfigWoddity( char *port ) {
	char **lp;
	LConfWoddity( NULL, tprintf( "!http %s", port ) );

	for (lp = _StandardConfigWoddity_Input; *lp; lp++) {
		LConfWoddity( NULL, tprintf( *lp ) );
	}
}

int InitVServers() {
	VServer_t	*VSPtr;
	int i;
	
	for (i=0; i<MAX_VSERVER; i++) {
		if ((VSPtr = VServers[i]) == NULL) break;
		
		if (verbose) printf( "Starting %s on %s\n", VSPtr->ServerName, VSPtr->Port );
		if (VSPtr->vsinit_fp) (VSPtr->vsinit_fp)( VSPtr );
	}
	return i;
}

