/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/


#include "woddity.h"

WService_t *NewWoddService( VServer_t *VSPtr, int prio ) {
	int n, i;
	WService_t	*wptr;
	
	if (VSPtr->NumWoddServices >= VSPtr->MaxWoddServices) {
		n = VSPtr->MaxWoddServices<<1;
		if (!n) n = 16;
		if ((wptr = realloc( VSPtr->WoddServices, sizeof( WService_t ) * n )) == NULL) return NULL;
		VSPtr->WoddServices = wptr;
		VSPtr->MaxWoddServices = n;
	}
	wptr = (VSPtr->WoddServices)+VSPtr->NumWoddServices;
	for (i=VSPtr->NumWoddServices; i>0; i--, wptr--) {
		if (((wptr-1)->flags & WS_PrioMASK) >= prio) break;
		memcpy( wptr, wptr-1, sizeof( WService_t ) );
	}
	VSPtr->NumWoddServices++;
	
	memset( wptr, 0, sizeof( WService_t ) );
	wptr->flags = WS_Valid | prio;
	return wptr;
}

WService_t *AttachHTTPService( VServer_t *VSPtr, ServiceCallback_t *sfp, char *_name, char *extra, ulong_t flags ) {
	WService_t	*wptr;
	char	*s, *t, *name, *ext=NULL;
	int	prio=0, match_mode=0;
	
	if (verbose & 4) printf( "AttachHTTPService \"%s\" to %s on %s\n", _name, VSPtr->ServerName, VSPtr->Port );

	name = strperm( _name );
	if (t = strchr( name, '*')) {		//  ?? * ??
		if (*name != '*') {			// ^ pre * ??
			prio += 4;
			match_mode |= WS_MatchPrefix;
		}
		*t++ = 0;
		if (*t) {			// ^ ?? * post $
			prio += 2;
			match_mode |= WS_MatchExt;
			ext = t;
		}
	} else {
		if (*name == '.') {
			prio = 1;			// ^ * .post $
			match_mode |= WS_MatchExt;
			ext = name;
		} else {
			prio = 8;			// ^ exact $
			match_mode |= WS_MatchName;
		}
	}
	
	if ((wptr = NewWoddService( VSPtr, prio )) == NULL) {
		return NULL;
	}
	wptr->flags |= flags | match_mode;
	if (name) {
		wptr->name = name;
		wptr->name_len = strlen( name );
	}
	if (ext) {
		wptr->ext = strperm( ext );
		wptr->ext_len = strlen( ext );
	}
	if (extra) wptr->extra = strperm( extra );

	wptr->fp = sfp;
	return wptr;
}

WService_t *MatchHTTPService( VServer_t *VSPtr, char *name ) {
	WService_t	*wptr;
	char			*s;
	int i, l, passnum;

	if (s = strchr( name, '*' )) *s=0;

	for (i=0, wptr = VSPtr->WoddServices; i<VSPtr->NumWoddServices; i++, wptr++) {
		if (wptr->flags & WS_MatchName) {
			if (strcmp( name, wptr->name ) != 0) continue;
		}
		if (wptr->flags & WS_MatchPrefix) {
			l = wptr->name_len;
			if (strncmp( name, wptr->name, l ) != 0) continue;
		}
		if (wptr->flags & WS_MatchExt) {
			l = wptr->ext_len;
			if (strcmp( name + strlen( name ) - l, wptr->ext ) != 0) continue;
		}
		return wptr;
	}
	return NULL;
}

int HTTPAlias( Channel_t *ChPtr, char *flnm, char *opts, char *extra ) {
	char *obuf, *d;
	char *post;
	char *s;
	int l;
	
	obuf = malloc( 1024 );
	ChannelAttachPTR( ChPtr, obuf );
	*obuf=0;

	if ((post = strchr( extra, '*' )) == NULL) {
		strcat( obuf, extra );
	} else {
		strncat( obuf, extra, post - extra );
		
		s = flnm;
		s+=ChPtr->wservptr->name_len;	// may be 0
		l = strlen( s );
		l-=ChPtr->wservptr->ext_len;	// may be 0
		strncat( obuf, s, l );
		if (*++post) strcat( obuf, post );
	}
	if (opts) {
		if (strchr( obuf, '?' )) {
			s = obuf + strlen( obuf ) -1;
			if (*s != '&') strcat( obuf, "&" );
		} else {
			strcat( obuf, "?" );
		}
		strcat( obuf, opts );
	}
	return DoHTTPService( ChPtr, obuf );
}

int DoHTTPService( Channel_t *ChPtr, char *s ) {
	WService_t	*wptr;
	char *opts;
	int r=0;
	
	SetRF( ChPtr, RF_Execute, RF_Read );
	if (*s == '/') s++;
	if (!*s) s = ChPtr->VSPtr->DefaultPage;
	
	if (opts = strchr( s, '?' )) *opts++ = 0;
	
	if (verbose & 2) printf( "DoHTTPService: >%s< opt=%s\n", s, (opts)?opts:"null" );
	if (verbose & 4) printf( "DoHTTPService %s on %s\n", ChPtr->VSPtr->ServerName, ChPtr->VSPtr->Port );

	TotalReqs++;
	if (wptr = MatchHTTPService( ChPtr->VSPtr, s )) {
//		if (wptr->flags & WS_MatchPrefix) s += wptr->name_len;
		
		if (wptr->root) ChPtr->root = wptr->root;
		else if (wptr->flags & WS_CGIRoot) ChPtr->root = ChPtr->VSPtr->CGIRoot;
		else if (wptr->flags & WS_USERRoot) ChPtr->root = ChPtr->VSPtr->USERRoot;
		else 	ChPtr->root = ChPtr->VSPtr->WWWRoot;


		ChPtr->wservptr = wptr;
		wptr->hit_count++;
		
		if (wptr->flags & WS_Thread) {
			r = ThreadHTTP( wptr, ChPtr, s, opts, wptr->extra );
		} else {
			r = (wptr->fp)( ChPtr, s, opts, wptr->extra );
		}
	} else r = 404;

	if (r) {
		bprintf( &(ChPtr->e), "Request: %s\n", s );
		MakeErrorPage( ChPtr, r );
	}
	return 0;
}

char *HTTP_status_str( int HTTP_status ) {
	char *stat_msg;
	switch (HTTP_status) {
//		case 11:	stat_msg = "CGI: SEGV";				break;
		case 200:	stat_msg = "OK";					break;
		case 201:	stat_msg = "Created";				break;
		case 202:	stat_msg = "Accepted";				break;
		case 204:	stat_msg = "No Content";			break;
		case 301:	stat_msg = "Moved Permanently";		break;
		case 302:	stat_msg = "Moved Temporarily";		break;
		case 304:	stat_msg = "Not Modified";			break;
		case 400:	stat_msg = "Bad Request";			break;
		case 401:	stat_msg = "Unauthorized";			break;
		case 403:	stat_msg = "Forbidden";				break;
		case 404:	stat_msg = "Not Found";				break;
		case 500:	stat_msg = "Internal Server Error";	break;
		case 501:	stat_msg = "Not Implemented";		break;
		case 502:	stat_msg = "Bad Gateway";			break;
		case 503:	stat_msg = "Service Unavailable";	break;
		case 999:	stat_msg = "Incomplete Request";	break;	// not HTTP
		default:	
			return tprintf( "Unknown Error: %d", HTTP_status );
	}
	return stat_msg;
}

#define	head_buffer_size	1024

int InsertHTTPHeadder( Channel_t *ChPtr, int HTTP_status ) {
	time_t	tv;
	char *s, *t, *q;
	int	n, sz, bsz;
//	char tbuf[1024];
	char	*tbptr;
	char *cont_type=NULL;
	char *cont_len=NULL;
	char *expires=NULL;
	Buffer_t	*bptr;

	bptr = &(ChPtr->o);
	
	if (!strncmp( _BufReadPtr( bptr ), "HTTP/1.", 6)) return 0;	// has headder

	if ((tbptr = ChPtr->head_buffer) == NULL) {
		tbptr = ChPtr->head_buffer = malloc( head_buffer_size );
	}
	
	bsz = min( head_buffer_size, BufferLength( bptr ) );

	memcpy( tbptr, _BufReadPtr( bptr ), bsz );
	s = tbptr;
//	if (verbose & 16) HexDump( s, 64 );
	while (1) {
		if ((t = strchr( s, '\n' )) == NULL) break;
		if ((t-s) == 0) break;
		if ((t-(tbptr)) >= bsz ) break;
		if (*(t-1) < 0x20) *(t-1)=0;
		*t++=0;
		if ((t-s) <= 2) break;
		if (*t == '\r') *t++=0;
		if (verbose&8) printf( "<CGI> %s\n", s );
		if ((q = strchr( s, ':' ))==NULL) break;
		*q++=0;
		q++;
		if (stricmp( s, "Content-type" ) == 0) {
			cont_type = strdup( q );
			if (verbose&8) printf( "<CGI> >%s<>%s<\n", s, q );
		} else if (stricmp( s, "Content-length" ) == 0) {
			cont_len = strdup( q );
			if (verbose&8) printf( "<CGI> >%s<>%s<\n", s, q );
//		} else if (strcmp( s, "Expires" ) == 0) {
//			expires = strdup( q );
		}
		s=t;
	}

	n = (s - tbptr) + 1;
	sz = BufferLength( bptr ) - n;
	if (verbose&8) printf( "<CGI> data size=%d  header=%d\n", sz, n );

	tv = time( NULL );
	strftime( tbptr, head_buffer_size, "%a, %d %b %Y %T", gmtime( &tv ) );
	biprintf( bptr, "Date: %s GMT\n", tbptr );
	biprintf( bptr, "Last-modified: %s GMT\n", tbptr );
	
	if (!cont_len) biprintf( bptr, "Content-length: %d\n", sz );
	if (!cont_type) biprintf( bptr, "Content-type: text/plain\n" );
	biprintf( bptr, "Server: %s\n", ServerSoftware );

	biprintf( bptr, "HTTP/1.0 %d %s\n", HTTP_status, HTTP_status_str( HTTP_status ) );
	
	if (cont_type) free( cont_type );
	if (cont_len) free( cont_len );
	if (expires) free( expires );
//	free( tbptr );
	return 0;
}

MakeErrorPage( Channel_t *ChPtr, int HTTP_status ) {
	time_t	tv;
	char	*stat_msg;
	int		sz, n;
//	char tbuf[4096];
	char	*tbptr;
	Buffer_t	*bptr;
	uchar_t 	*p;

	bptr = &(ChPtr->o);
	
	if ((tbptr = ChPtr->head_buffer) == NULL) {
		tbptr = ChPtr->head_buffer = malloc( head_buffer_size );
	}
	
	ClearBuffer( bptr );
	
	stat_msg = HTTP_status_str( HTTP_status );
	
	log_error( ChPtr, stat_msg );

	if ((HTTP_status < 200) || (HTTP_status >= 999)) HTTP_status = 500;
	
	bprintf( bptr, "<html>\n<head><title>%d - %s</title></head>\n"
					"<body bgcolor=\"#FFFFFF\" text=\"#660000\">\n"
					"<h1>%d - %s</h1>\n",
					HTTP_status, stat_msg,
					HTTP_status, stat_msg );
					
	if (verbose && (n = BufferLength( &(ChPtr->e) )) ) {
		bprintf( bptr, "<pre>\n" );
		
		BufferEnqueue( bptr, BufferDequeue( &(ChPtr->e), n ), n );
		bprintf( bptr, "</pre>\n" );
	}

	bprintf( bptr, "<p><b>%s</b><br>\n", ServerSoftware );
	tv = time( NULL );
	strftime( tbptr, head_buffer_size, "%a, %d %b %Y %T", gmtime( &tv ) );
	bprintf( bptr, "Date: %s GMT\n", tbptr );

	bprintf( bptr, "</body></html>\n\n" );
	
	sz = BufferLength( bptr );
						
	biprintf( bptr, "\n" );
	tv = time( NULL );
	strftime( tbptr, head_buffer_size, "%a, %d %b %Y %T", gmtime( &tv ) );
	biprintf( bptr, "Last-modified: %s GMT\n", tbptr );
	biprintf( bptr, "Date: %s GMT\n", tbptr );

	biprintf( bptr, "Content-length: %d\n", sz );
	biprintf( bptr, "Content-type: text/html\n" );
	biprintf( bptr, "Server: %s\n", ServerSoftware );

	biprintf( bptr, "HTTP/1.0 %d %s\n", HTTP_status, stat_msg );

	ChPtr->req_status = HTTP_status;
	SetRF( ChPtr, RF_Write, RF_Execute );
//	free( tbptr );
	return 0;
}

#define	ContentLength "Content-length:"

int ServeHTTPRequest( Channel_t *ChPtr ) {
	time_t	tv;
	char	time_buf[64];
	char tbuf[256];
	char *reqt, *cmd;
	char *s, *d, *ln, *q;
	char *content;
	int	HeadSize;
	int	CurContentLength=0;
	int	HeadContentLength=0;
	Buffer_t	*bptr;
	
	bptr = &(ChPtr->i);
	
	s = _BufReadPtr( bptr );
	if ((d = strchr( s, 0x0A )) == NULL) return -1;
	memset( tbuf, 0, sizeof( tbuf ) );
	strncpy( tbuf, s, min( (d-s), sizeof( tbuf ) ) );
	
	if ((d = strstr( s, "\x0D\x0A\x0D\x0A" )) != NULL) {
		content = d+4;
	} else if ((d = strstr( s, "\x0A\x0A" )) != NULL) {
		content = d+2;
	}

	if (d == NULL) {
		if (verbose&2) printf( "*** Incomplete Headder ***\n" );
		return 0;
	}
	HeadSize = (content - s);
	CurContentLength = _BufSize( bptr ) - HeadSize;
	ChPtr->content_offset = HeadSize;
	
	if (s = strstr( _BufReadPtr( bptr ), ContentLength )) {
		s += strlen( ContentLength );
		while (*s && (*s >= 0x20) && !isdigit( *s )) s++;
		if (isdigit( *s )) {
			HeadContentLength = strtoul( s, NULL, 0 );
			ChPtr->content_length = HeadContentLength;
		}
		if (verbose) printf( "Content-Length: %d of %d\n", CurContentLength, HeadContentLength );
		if (CurContentLength < HeadContentLength) {
			if (verbose&2) printf( "*** Incomplete Content ***\n" );
			return 0;
		}
	}

	if (verbose&2) printf( "REQUEST: %s\n", tbuf );
	
	s = tbuf;
	while (*s) {
		while (*s && (*s <= 0x20)) s++;
		reqt = s;
		while (*s && (*s > 0x20)) s++;
		*s++ = 0;
		
		while (*s && (*s <= 0x20)) s++;
		cmd = s;
		while (*s && (*s > 0x20)) s++;
		*s++ = 0;

		ChPtr->req_status = 999;
		q=inet_ntoa( ChPtr->addr.sin_addr );
		tv = time( NULL );
		strftime( time_buf, sizeof( time_buf ), "%d/%b/%Y:%H:%M:%S GMT", gmtime( &tv ) );
		if (ChPtr->VSPtr) {
			sprintf( ChPtr->req_log, "%s - - [%s] \"%s %s:%s%s HTTP/1.0\"", q, time_buf, 
				reqt, ChPtr->VSPtr->ServerName, ChPtr->VSPtr->Port, cmd );
		} else {
			sprintf( ChPtr->req_log, "%s - - [%s] \"%s %s HTTP/1.0\"", q, time_buf, reqt, cmd );
		}
		
#if EnableVisitorTrack
		RegVisitorTrack( ChPtr, ChPtr->addr.sin_addr );
#endif
			
		if (stricmp( reqt, "GET" ) == 0) {
			ChPtr->serve_type = ServeGET;
			return DoHTTPService( ChPtr, cmd );
		}
		if (stricmp( reqt, "HEAD" ) == 0) {
			ChPtr->serve_type = ServeHEAD;
			return DoHTTPService( ChPtr, cmd );
		}
		if (stricmp( reqt, "POST" ) == 0) {
			ChPtr->serve_type = ServePOST;
			return DoHTTPService( ChPtr, cmd );
		}
		if (verbose) {
			printf( "*** Unknown request ***\n" );
			HexDump( _BufReadPtr( &(ChPtr->i )), _BufSize( &(ChPtr->i )) );
		}
		break;
	}
	return 0;
}

