/*--------------------------------------------------------------------------

    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"

int SocketRead( Buffer_t *bptr, int fd, int free_space ) {
	char	*p;
	int n, r;

	for (r=0; r<3; r++) {
		p = BufferMakeSpace( bptr, READBUF_MIN );			
		if ((n = read( fd, p, BufferFreeSpace( bptr ))) <= 0) {
			if (errno == EINTR) continue;
			return n;
		}
		BufferEnqueue( bptr, NULL, n );
		return n;
	}
	return -1;
}

unsigned last_CDtime=0;

int ChEventHTTP_TCP( Channel_t *ChPtr, unsigned chev, int stat ) {
	int n, r, t, retry;
	Buffer_t	*bptr;
	char	*p;

	switch (chev) {
		case CHEV_Read:
			if (!ChPtr->fdr) return 0;
			
			if (ChPtr->SPtr) bptr = &(ChPtr->SPtr->o);
			else bptr = &(ChPtr->i);
			
			if ((n = SocketRead( bptr, ChPtr->fdr, READBUF_MIN )) <= 0) {
				RFDclose( ChPtr->fdr );
				ChPtr->fdr = 0;
				SetRF( ChPtr, 0, RF_Read );
				return 0;
			}
			
			if (verbose & 8) printf( "Read %d   %d bytes\n", ChPtr->idx, n );

			SetRF( ChPtr, RF_Query, 0);
			break;
		case CHEV_Write: 
			if (ChPtr->SPtr) bptr = &(ChPtr->SPtr->i);
			else bptr = &(ChPtr->o);
			
			for (retry=0; ; retry++) {
				if ((n = BufferLength( bptr )) > WRITE_MAX) n = WRITE_MAX;
				if (n == 0) return 0;
				r = write( ChPtr->fdw, BufferDequeue( bptr, 0), n );
				if (r >= 0) break;
				if (errno == EINTR) {
					if (retry >= 5) return 0;
					continue;
				}
				if (verbose & 2) printf( "ERROR in Write %d   rslt %d, errno %d\n", ChPtr->idx, r, errno );
				RFDclose( ChPtr->fdw );
				ChPtr->fdw = 0;
				SetRF( ChPtr, 0, RF_Write );
				return 0;
			}
						
			if (verbose & 8) printf( "Write %d   %d bytes - result: %d\n", ChPtr->idx, n, r );
			if (r > 0) {
				BufferDequeue( bptr, r );
				ChPtr->total_write += r;
				if ((!TestRF( ChPtr, RF_Execute)) && (BufferLength( bptr ) == 0)) {
					SetRF( ChPtr, 0, RF_Write );
					ChannelDistroy( ChPtr, 0 );
				}
			}
			break;
		case CHEV_ReadE: 
			if (ChPtr->SPtr) bptr = &(ChPtr->SPtr->e);
			else bptr = &(ChPtr->e);
	
			if ((n = SocketRead( bptr, ChPtr->fde, READBUF_MIN )) <= 0) {
				RFDclose( ChPtr->fde );
				ChPtr->fde = 0;
				return 0;
			}
			log_error( ChPtr, _BufWritePtr( bptr ) - n );
			break;
		case CHEV_Query: 
			ServeHTTPRequest( ChPtr );
			break;
		case CHEV_Expire: 
			if (ChPtr->thread_pid) kill( ChPtr->thread_pid, SIGINT );
			else ChannelDistroy( ChPtr, 0 );
			break;
		case CHEV_Init: 
			ChPtr->expire = ChPtr->VSPtr->ChildExpire*1000;
			break;
		case CHEV_Child:
			SetRF( ChPtr, 0, RF_Read );
			if (TestRF( ChPtr, RF_Child)) break;
			if (TestRF( ChPtr, RF_Execute)) {
				if (BufferLength( &(ChPtr->o) ) == 0) {
					MakeErrorPage( ChPtr, 500 );
				}
				SetRF( ChPtr, RF_Write, 0 );
			} 
			break;
		case CHEV_Close: 
			shutdown( ChPtr->fdr, 0 );
			AddMetric( &metric_xfer, ChPtr->total_write );
			if (ChPtr->create_time) {
				AddMetric( &metric_time, ChPtr->total_time );
			}
			if (ChPtr->VTPtr) {		// per visitor metrics
				if (ChPtr->total_time) AddMetric( &(ChPtr->VTPtr->metric_time), ChPtr->total_time );
				if (ChPtr->total_write) AddMetric( &(ChPtr->VTPtr->metric_xfer), ChPtr->total_write );
			}
			if (ChPtr->wservptr) {		// per rule metrics
				if (ChPtr->total_time) AddMetric( &(ChPtr->wservptr->metric_time), ChPtr->total_time );
				if (ChPtr->total_write) AddMetric( &(ChPtr->wservptr->metric_xfer), ChPtr->total_write );
			}
			if (ChPtr->VSPtr) {		// per VServer metrics
				if (ChPtr->total_time) AddMetric( &(ChPtr->VSPtr->metric_time), ChPtr->total_time );
				if (ChPtr->total_write) AddMetric( &(ChPtr->VSPtr->metric_xfer), ChPtr->total_write );
				ChPtr->VSPtr->hit_count++;
			}
			if (last_CDtime) {
				t = GetCurrentTime() - last_CDtime;
				AddMetric( &metric_qrate, t );
			} 
			last_CDtime = GetCurrentTime();
			log_req_status( ChPtr );
			
			break;
	}
	return 0;
}

int ChEventHTTP_PIPE( Channel_t *ChPtr, unsigned chev, int stat ) {
	int n;
	Buffer_t	*bptr;

	switch (chev) {
		case CHEV_Read: 
		case CHEV_Write: 
		case CHEV_ReadE: 
		case CHEV_Expire: 
			return ChEventHTTP_TCP( ChPtr, chev, stat );

		case CHEV_Query: 
			break;
		case CHEV_Init: 
			ChPtr->expire = 2*60*1000;
			break;
		case CHEV_Child:
			SetRF( ChPtr, 0, RF_Execute );
			if (ChPtr->SPtr) ChPtr->SPtr->req_status = stat;
			ChPtr->req_status = stat;
			break;
		case CHEV_Close: {
			Channel_t	*PChPtr;
			
			if (PChPtr = ChPtr->SPtr) {
				PChPtr->ChildPtr = NULL;
				SetRF( PChPtr, 0, RF_Child );
				if (PChPtr->cxcb_fp) {
					(PChPtr->cxcb_fp)( PChPtr, ChPtr->req_status );
				} else {
					SetRF( ChPtr, RF_Write, RF_Execute );
				}
			}
			ChPtr->SPtr = NULL;
			break;
		}
	}
	return 0;
}

int ChEventHTTP_Listen( Channel_t *ChPtr, unsigned chev, int stat ) {
	Channel_t	*NChPtr;
	int retry;

	if (ChPtr->VSPtr == NULL) return 0;
	
	switch (chev) {
		case CHEV_Read: {
			NChPtr = ChannelCreate( ChPtr, ChType_TCP, ChEventHTTP_TCP );
			NChPtr->addr_len = sizeof( NChPtr->addr );
			for (retry=0; ; retry++) {
				if ((NChPtr->fdr = NChPtr->fdw = accept( ChPtr->fdr, (struct sockaddr *)&(NChPtr->addr), &(NChPtr->addr_len) )) >= 0) break;
				if (errno != EINTR) return 0;
				if (retry >= 3) return 0;
			}
			NChPtr->SPtr = NULL;
			RegFD( NChPtr->fdr, RFD_open | RFD_openTCP, NChPtr->idx );
			sprintf( NChPtr->req_log, "accept %d", NChPtr->fdr );
			if (verbose & 4) printf( "Accepting\n" );
			break;
		}
		case CHEV_Init: {
			int opt_v;
			
			if ((ChPtr->fdr = socket( AF_INET, SOCK_STREAM, 0 )) < 0) {
				perror( "opening listen socket" );
				exit( 1 );
			}
			if (verbose > 2) printf( "Listener as %d\n", ChPtr->fdr );

			RegFD( ChPtr->fdr, RFD_open | RFD_openTCP, ChPtr->idx );
			
			opt_v = 1;
			setsockopt( ChPtr->fdr, SOL_SOCKET, SO_REUSEADDR, &opt_v, sizeof( opt_v ) );

			ChPtr->addr.sin_family = AF_INET;
			if (ChPtr->VSPtr->flags & VS_LOCALONLY) {
				ChPtr->addr.sin_addr.s_addr = ntohl( INADDR_LOOPBACK );
			} else {
				ChPtr->addr.sin_addr.s_addr = INADDR_ANY;
			}
			ChPtr->addr.sin_port = htons( strtol( ChPtr->VSPtr->Port, NULL, 0 ) );
	
			if (bind( ChPtr->fdr, (struct sockaddr *)&(ChPtr->addr), sizeof( ChPtr->addr ) ) ) {
				perror( "binding listen socket" );
				exit( 1 );
			}
	
			ChPtr->addr_len = sizeof( ChPtr->addr );
			if (getsockname( ChPtr->fdr, (struct sockaddr *)&(ChPtr->addr), &(ChPtr->addr_len) ) ) {
				perror( "getting socket name" );
				return -1;
			}
	
			sprintf( ChPtr->req_log, "listen port %s", ChPtr->VSPtr->Port );
	
			listen( ChPtr->fdr, 64 );
			FreeBuffer( &(ChPtr->i) );
			FreeBuffer( &(ChPtr->o) );
			FreeBuffer( &(ChPtr->e) );
			break;
		}
		case CHEV_Close: {
			int addr_len;
			struct sockaddr_in	addr;
			int h;

			if (ChPtr->fdr) {
				shutdown( ChPtr->fdr, 1 );
				if (ChPtr->req_status++ > 3) {
					RFDclose( ChPtr->fdr );
					ChPtr->fdr = ChPtr->fdw = 0;
				}
				sprintf( ChPtr->req_log, "Shutdown %d", ChPtr->req_status );
				return -1;
			}
#if 0
			
			addr_len = sizeof( addr );
			while ((h = accept( ChPtr->fdr, (struct sockaddr *)&addr, &addr_len )) >= 0) {
				RegFD( h, RFD_open | RFD_openTCP, ChPtr->idx );
//				shutdown( h, 2 );
				RFDclose( h );
			}
			RFDclose( ChPtr->fdr );
			ChPtr->fdr = ChPtr->fdw = 0;
#endif
			
			break;
		}
		default:	return -1;
	}
	return 0;
}

int AttachHTTP_Listen( VServer_t *VSPtr ) {
	Channel_t *ChPtr;
	
	ChPtr = ChannelCreate( NULL, ChType_Listen, ChEventHTTP_Listen );
	ChPtr->VSPtr = VSPtr;
	ChannelEvent( ChPtr, CHEV_Init, 0 );
	

	return 0;
}
