/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"
#ifdef __QNX__
#include "sys/sched.h"
#else
#include "sched.h"
#endif

extern char **environ;


int SetCloseOnExec( int fd ) {
	return fcntl( fd, F_SETFD, fcntl( fd, F_GETFD ) | FD_CLOEXEC );
}

int envprintf( Channel_t *ChPtr, const char *frmt, ... ) {
	va_list	arglist;
	int l;
	char *p;
	
	if (ChPtr->env == NULL) {
		ChPtr->env = malloc( MAX_ENV_BUFFER );
		malloc_lock( ChPtr->env, 1 );
	}
	if (ChPtr->encl == 0) {
		ChPtr->encl = MAX_ENV * sizeof( char * );
	}
	if (ChPtr->enc >= MAX_ENV) return -1;
	p = ((char *)(ChPtr->env)) + ChPtr->encl;
	ChPtr->env[ChPtr->enc++] = p;
	ChPtr->env[ChPtr->enc] = NULL;
	
	va_start( arglist, frmt );
	
	l = vsnprintf( p, MAX_ENV_BUFFER - ChPtr->encl, frmt, arglist );
	ChPtr->encl += l + 1;

	return l;
}

int argprintf( Channel_t *ChPtr, const char *frmt, ... ) {
	va_list	arglist;
	int l;
	char *p;
	
	if (ChPtr->argv == NULL) {
		ChPtr->argv = malloc( MAX_ARGV_BUFFER );
		malloc_lock( ChPtr->argv, 1 );
	}
	if (ChPtr->argcl == 0) {
		ChPtr->argcl = MAX_ARGV * sizeof( char * );
	}
	if (ChPtr->argc >= MAX_ARGV) return -1;
	p = ((char *)(ChPtr->argv)) + ChPtr->argcl;
	ChPtr->argv[ChPtr->argc++] = p;
	ChPtr->argv[ChPtr->argc] = NULL;
	
	va_start( arglist, frmt );
	
	l = vsnprintf( p, MAX_ARGV_BUFFER - ChPtr->argcl, frmt, arglist );
	ChPtr->argcl += l + 1;

	return l;
}

int BuildENVandARGV( Channel_t *PChPtr, Channel_t *ChPtr, char *flnm, char *opts, char *extra ) {
	char *q, *s, *t, *bp, *d, *tflnm;
	char tbuf[CGIBUF_SIZE];
	int enc=0;
	
	tflnm = flnm;

	if (ChPtr->wservptr->alias) {
		tflnm = ChPtr->wservptr->alias;
	} else {
		if (ChPtr->wservptr && (ChPtr->wservptr->flags & WS_MatchPrefix)) {
			tflnm += ChPtr->wservptr->name_len;
		}
	}
	
	if (opts) envprintf( PChPtr, "QUERY_STRING=%s", opts );

	if (ChPtr->wservptr && (ChPtr->wservptr->flags & WS_MatchPrefix)) {
		envprintf( PChPtr, "PATH_INFO=/%s", tflnm );
		envprintf( PChPtr, "PATH_TRANSLATED=%s/%s", ChPtr->VSPtr->WWWRoot, tflnm );
	}

	envprintf( PChPtr, "DOCUMENT_ROOT=%s", ChPtr->VSPtr->WWWRoot );
	envprintf( PChPtr, "GATEWAY_INTERFACE=CGI" );
	envprintf( PChPtr, "HOME=%s", ChPtr->root );
	envprintf( PChPtr, "PATH=%s:/bin:/usr/bin", ChPtr->root );
	envprintf( PChPtr, "PWD=%s", ChPtr->root );
	envprintf( PChPtr, "SCRIPT_FILENAME=%s/%s", ChPtr->root, tflnm );
	envprintf( PChPtr, "SCRIPT_NAME=/%s", flnm );
	envprintf( PChPtr, "SERVER_ADMIN=%s", ChPtr->VSPtr->ServerAdmin );
	envprintf( PChPtr, "SERVER_NAME=%s", ChPtr->VSPtr->ServerName );
	envprintf( PChPtr, "SERVER_PORT=%s", ChPtr->VSPtr->Port );
	envprintf( PChPtr, "SERVER_PROTOCOL=HTTP/1.0" );
	envprintf( PChPtr, "SERVER_SOFTWARE=%s", ServerSoftware );
	envprintf( PChPtr, "REMOTE_PORT=%d", ntohs( ChPtr->addr.sin_port ) );
	q=inet_ntoa( ChPtr->addr.sin_addr );
	envprintf( PChPtr, "REMOTE_ADDR=%s", q );
	
	switch (ChPtr->serve_type) {
		case ServeGET:	s = "GET";	break;
		case ServePOST:	s = "POST";	break;
		default:		s = "unknown";	
	}
	envprintf( PChPtr, "REQUEST_METHOD=%s", s );

	if (verbose & 4) envprintf( PChPtr, "__FLNM=%s", flnm );
	if (verbose) envprintf( PChPtr, "__TotalChildren=%d", TotalChildren );
	if (verbose) envprintf( PChPtr, "__content_length=%d", ChPtr->content_length );
	
	bp = s = _BufReadPtr( &(ChPtr->i ) );
	while (1) {
		if ((t = strchr( s, '\n' )) == NULL) break;
		if ((t-s) == 0) break;
		if ((t-bp) >= _BufSize( &(ChPtr->i) )) break;
		if (*(t-1) < 0x20) *(t-1)=0;
		*t++=0;
		if ((t-s) <= 2) break;
		if (*t == '\r') *t++=0;
		if ((q = strchr( s, ':' ))!=NULL) {
			*q++=0;
			q++;
			for (d = tbuf; *s; s++) {
				switch (*s) {
					case '-':	*d++ = '_';	break;
					default:	*d++ = toupper( *s );
				}
				*d = 0;
			}
	
			envprintf( PChPtr, "HTTP_%s=%s", tbuf, q );
		}
		s=t;
	}

	argprintf( PChPtr, "%s/%s", ChPtr->root, tflnm );
	s=opts;
	while (s && *s) {
		t=s;
		while (*t && (*t != '&')) t++;
		if (*t) *t++ = 0;
		argprintf( PChPtr, "%s", s );
		s=t;
	}
	return 0;	
}
	

WService_t *AttachHTTPBuiltin( VServer_t *VSPtr, BuiltinCallback_t *xfp, char *name, char *extra, ulong_t flags ) {
	WService_t	*wptr;

	if ((wptr = AttachHTTPService( VSPtr, HTTPBuiltin, name, extra, flags )) == NULL) return NULL;
	wptr->xfp = xfp;

	return wptr;
}

int HTTPExecExitCB( Channel_t *ChPtr, int stat ) {
	if (verbose & 2) printf( "HTTPExecExitCB %d - %s\n", stat, ChPtr->req_log );
	if (stat) {
		bprintf( &(ChPtr->e), "ERROR %d Request: %s\n", stat, ChPtr->req_log );
		MakeErrorPage( ChPtr, stat );
	} else {
		ChPtr->req_status = 200;
		InsertHTTPHeadder( ChPtr, ChPtr->req_status );
		
		SetRF( ChPtr, RF_Write, RF_Execute );
	}

	SendKICK( "ExecCB " );

	return 0;
}

int HTTPBuiltin( Channel_t *ChPtr, char *flnm, char *opts, char *extra ) {
	int argc=0, enc=0, r, i, pipeI[2], pipeO[2], pipeE[2], pid, n, pI, pO, pE;
	char *cmd;
	char tbuf[CGIBUF_SIZE];
	char *q, *s, *t, *bp, *d;
	int rslt=500, stat=0, child_status;
	Channel_t	*PChPtr;

	if ((ChPtr->wservptr == NULL) || (ChPtr->wservptr->xfp == NULL)) return 502;

	while (1) {
		pipeI[0] = pipeI[1] = 
		pipeO[0] = pipeO[1] = 
		pipeE[0] = pipeE[1] = -1;

		PChPtr = ChannelCreate( ChPtr, ChType_PIPE, ChEventHTTP_PIPE );
		PChPtr->SPtr = ChPtr;
		ChPtr->ChildPtr = PChPtr;
		SetRF( ChPtr, RF_Child, 0 );
		ChPtr->cxcb_fp = HTTPExecExitCB;
		
		BuildENVandARGV( PChPtr, ChPtr, flnm, opts, extra );	
		
		sprintf( PChPtr->req_log, "%s --> %s", ChPtr->req_log, PChPtr->argv[0] );

		if (ChPtr->content_length) {
			if (pipe( pipeI ) < 0) break;
			PChPtr->fdw = pipeI[1];
			SetCloseOnExec( pipeI[1] );
			BufferEnqueue( &(ChPtr->i), "\n", 1 );
			ChPtr->content_length++;
			ChPtr->i.read = ChPtr->i.write - ChPtr->content_length;
			SetRF( PChPtr, RF_Write, 0 );
			
			RegFD( pipeI[0], RFD_open | RFD_openPIPE0, PChPtr->idx );
			RegFD( pipeI[1], RFD_open | RFD_openPIPE1c, PChPtr->idx );
		}
		if (pipe( pipeO ) < 0) break;
		PChPtr->fdr = pipeO[0];
		SetRF( PChPtr, RF_Read, 0 );
		RegFD( pipeO[0], RFD_open | RFD_openPIPE0c, PChPtr->idx );
		RegFD( pipeO[1], RFD_open | RFD_openPIPE1, PChPtr->idx );
		
		if (pipe( pipeE ) < 0) break;
		PChPtr->fde = pipeE[0];
		RegFD( pipeE[0], RFD_open | RFD_openPIPE0c, PChPtr->idx );
		RegFD( pipeE[1], RFD_open | RFD_openPIPE1, PChPtr->idx );

		pid = fork ();
		if (pid == (pid_t) -1) {		// fork failed - cleanup
			rslt=503;
			break;
		} else if (pid == 0) {			// fork child
			while (1) {
				sched_yield();
				if (pipeI[0] >= 0) dup2( pipeI[0], 0 );
				else close( 0 );
				dup2( pipeO[1], 1 );
				dup2( pipeE[1], 2 );
				chdir( ChPtr->root );
				
//				environ = PChPtr->env;
				
				r = (ChPtr->wservptr->xfp)( PChPtr->argc-1, PChPtr->argv, PChPtr->env );
				_exit( r );
			}
			_exit( 127 );
		} else {					// fork parent
			if (verbose&2) printf( "*** fork %d\n", pid );
			TotalChildren++;
			if (TotalChildren > MaxTotalChildren) MaxTotalChildren = TotalChildren;
			SetRF( PChPtr, RF_Execute, 0 );
			PChPtr->thread_pid = pid;
			if (pipeI[0] >= 0) RFDclose( pipeI[0] );
			RFDclose( pipeO[1] );
			RFDclose( pipeE[1] );
			return 0;
		}
	}
			
	if (pipeI[0]>=0) RFDclose( pipeI[0] );
	if (pipeO[1]>=0) RFDclose( pipeO[1] );
	if (pipeE[1]>=0) RFDclose( pipeE[1] );
	if (PChPtr) ChannelDistroy( PChPtr, 1 );

	return rslt;
}

int HTTPExec( Channel_t *ChPtr, char *flnm, char *opts, char *extra ) {
	char iov[10];
	int argc=0, enc=0, r, i, pipeI[2], pipeO[2], pipeE[2], pid, n, pI, pO, pE;
	char *cmd;
	char tbuf[CGIBUF_SIZE];
	char *q, *s, *t, *bp, *d;
	int rslt=500, stat=0, child_status;
	Channel_t	*PChPtr;
	
	while (1) {
		memset( iov, 0xFF, sizeof( iov ) );
		pipeI[0] = pipeI[1] = 
		pipeO[0] = pipeO[1] = 
		pipeE[0] = pipeE[1] = -1;

		PChPtr = ChannelCreate( ChPtr, ChType_PIPE, ChEventHTTP_PIPE );
		PChPtr->SPtr = ChPtr;
		ChPtr->ChildPtr = PChPtr;
		SetRF( ChPtr, RF_Child, 0 );
		ChPtr->cxcb_fp = HTTPExecExitCB;
		
		BuildENVandARGV( PChPtr, ChPtr, flnm, opts, extra );	
		
		strcpy( PChPtr->req_log, PChPtr->argv[0] );

		if (ChPtr->content_length) {
			if (pipe( pipeI ) < 0) break;
			PChPtr->fdw = pipeI[1];
			iov[0] = pipeI[0];
			SetCloseOnExec( pipeI[1] );
			BufferEnqueue( &(ChPtr->i), "\n", 1 );
			ChPtr->content_length++;
			ChPtr->i.read = ChPtr->i.write - ChPtr->content_length;
			SetRF( PChPtr, RF_Write, 0 );
			
			RegFD( pipeI[0], RFD_open | RFD_openPIPE0, PChPtr->idx );
			RegFD( pipeI[1], RFD_open | RFD_openPIPE1c, PChPtr->idx );
		}
		if (pipe( pipeO ) < 0) break;
		iov[1] = pipeO[1];
		PChPtr->fdr = pipeO[0];
		SetRF( PChPtr, RF_Read, 0 );
		RegFD( pipeO[0], RFD_open | RFD_openPIPE0c, PChPtr->idx );
		RegFD( pipeO[1], RFD_open | RFD_openPIPE1, PChPtr->idx );
		
		if (pipe( pipeE ) < 0) break;
		iov[2] = pipeE[1];
		PChPtr->fde = pipeE[0];
		RegFD( pipeE[0], RFD_open | RFD_openPIPE0c, PChPtr->idx );
		RegFD( pipeE[1], RFD_open | RFD_openPIPE1, PChPtr->idx );


#ifdef XXX__QNX__
		if ((pid = qnx_spawn( 0, NULL, 0, -1, -1, 0, PChPtr->argv[0], PChPtr->argv, PChPtr->env, iov, -1 )) == -1) {
			fprintf( stderr, "*** qnx_spawn failed: %s\n", strerror( errno ) );
			Cprintf( ChPtr, ChannelEEnqueue, "Spawn Failed: %s\n", strerror( errno ) );
			rslt = 404;
			break;
		}
#endif

#ifdef __QNX__
		pid = fork ();
#else
		pid = __vfork ();
#endif
		if (pid == (pid_t) -1) {		// fork failed - cleanup
			rslt=503;
			break;
		} else if (pid == 0) {			// fork child
			while (1) {
				sched_yield();
				if (pipeI[0] >= 0) dup2( pipeI[0], 0 );
				else close( 0 );
				dup2( pipeO[1], 1 );
				dup2( pipeE[1], 2 );
				chdir( ChPtr->root );
				execve( PChPtr->argv[0], PChPtr->argv, PChPtr->env );
				_exit( 127 );
			}
			_exit( 127 );
		} else {					// fork parent
			if (verbose&2) printf( "*** vfork %d\n", pid );
			TotalChildren++;
			if (TotalChildren > MaxTotalChildren) MaxTotalChildren = TotalChildren;
			SetRF( PChPtr, RF_Execute, 0 );
			PChPtr->thread_pid = pid;
			if (pipeI[0] >= 0) RFDclose( pipeI[0] );
			RFDclose( pipeO[1] );
			RFDclose( pipeE[1] );
			return 0;
		}
	}
			
	if (pipeI[0]>=0) RFDclose( pipeI[0] );
	if (pipeO[1]>=0) RFDclose( pipeO[1] );
	if (pipeE[1]>=0) RFDclose( pipeE[1] );
	if (PChPtr) ChannelDistroy( PChPtr, 1 );

	return rslt;
}

int Thread_RunService( void *args[] ) {
	WService_t	*wptr;
	Channel_t	*ChPtr;
	FILE *fp;
	char *flnm, *opts, *whatever;
	int	r;
	
	wptr = args[0];
	ChPtr = args[1];
	flnm = args[2];
	opts = args[3];
	whatever = args[4];
	
	r = (wptr->fp)( ChPtr, flnm, opts, whatever );
	if (r == 200) r = 0;
	if (r) {
		if (r < 0) r = errno;
		bprintf( &(ChPtr->e), "Request: %s%c%s  Error: %d\n", flnm, (opts)?'?':' ', (opts)?opts:"" );
		MakeErrorPage( ChPtr, r );
	}
#ifdef __QNX__
	Trigger( kick_proxy );
 	_endthread();
#else
	SendKICK( "Thread " );
	exit( 0 );
#endif
	return 0;
}

int ThreadHTTP( WService_t *wptr, Channel_t *ChPtr, char *flnm, char *opts, char *whatever ) {
	void	**args;
	int	pid, m;

	if (MaxChildren && (TotalChildren >= MaxChildren)) return 503;
	TotalChildren++;
	if (TotalChildren > MaxTotalChildren) MaxTotalChildren = TotalChildren;
	if (ChPtr->STACK_p == NULL) {
		ChPtr->STACK_max = STACK_SIZE;
		ChPtr->STACK_p = malloc( ChPtr->STACK_max );
		if ((m = malloc_size( ChPtr->STACK_p )) > ChPtr->STACK_max) ChPtr->STACK_max = m;
		malloc_lock( ChPtr->STACK_p, 1 );
	} 
	
	memset( ChPtr->STACK_p, 0, ChPtr->STACK_max );
	
	args = malloc( 32 );

	args[0] = wptr;
	args[1] = ChPtr;
	args[2] = strdup( flnm );
	args[3] = strdup( (opts) ? opts : "" );
	args[4] = strdup( (whatever) ? whatever : "" );
	args[5] = NULL;

	ChannelAttachPTR( ChPtr, args );
	ChannelAttachPTR( ChPtr, args[2] );
	ChannelAttachPTR( ChPtr, args[3] );
	ChannelAttachPTR( ChPtr, args[4] );

#ifdef __QNX__
	pid = _beginthread( Thread_RunService, ChPtr->STACK_p, ChPtr->STACK_max, args );
#else
	pid = __clone( Thread_RunService, ChPtr->STACK_p+ChPtr->STACK_max-64,
		SIGCHLD | CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND , args );
#endif	
	if (pid == -1) return 503;
	if (verbose&2) printf( "*** clone %d\n", pid );
	ChPtr->thread_pid = pid;
	
	return 0;
}

