/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/select.h>
#include <malloc.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <arpa/inet.h>

#include <time.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <sys/wait.h>

#ifdef __QNX__
#include <ioctl.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#endif

#include "woddmem.h"
#include "woddfd.h"
#include "woddoptions.h"

#ifndef __Woddity__Included__
#define __Woddity__Included__

#ifdef __cplusplus
extern "C" {
#endif

#ifndef	WEXT
#define	WEXT	extern
#define	WINIT( x )
#else	
#define	WINIT(x) =x
#endif

#ifndef ulong_t
typedef unsigned long ulong_t;
#endif
#ifndef ushort_t
typedef unsigned short ushort_t;
#endif
#ifndef uchar_t
typedef unsigned char uchar_t;
#endif
#ifndef min
#define min( a, b ) (((a)<(b))?(a):(b))
#endif

#define	WTSecond		1000
#define	WTMinute		(WTSecond*60)
#define	WTHour			(WTMinute*60)
#define	WTDay			(WTHour*24)

#define	WRITE_MAX		0x2000	
#define	READFILE_MAX		0x4000 
#define	READFILE_BUFLIMIT	0x100	// 0x1000	// should keep buffers to 64K
#define	WRITEPOST_MAX	4096
#define	CGIBUF_SIZE		256
#define	CGIBUF_MIN		4096
#define	READBUF_MIN		4096
#define STACK_SIZE		(6000)
#define	MAX_ENV			128
#define	MAX_ARGV		64
#define	MAX_ENV_BUFFER			8192
#define	MAX_ARGV_BUFFER		4096
#define	MAX_VSERVER	64

#define	CHANNEL_ATROPHY		(WTSecond * 60)

#define	WODD_VER_MAJ		0
#define	WODD_VER_MIN		9
#define	WODD_VER_REV		"b"

WEXT char *ServerSoftware;		// assigned in RunWoddity
WEXT char *WoddProgName;		// assigned in WoddityMain

WEXT int TotalChildren;
WEXT int MaxTotalChildren;
WEXT int MaxChildren WINIT( 64 );
WEXT ulong_t verbose WINIT( 1 );
WEXT int MainPID;
WEXT int InSelect;
WEXT char *LOGRoot WINIT( "/var/log/woddity" );

WEXT int StartTime;
WEXT int RunDays;

WEXT char *LogName;
WEXT int LogHandle WINIT( -1 );

WEXT unsigned	RF_Changed;
WEXT pid_t	kick_proxy;
WEXT int kick_fd;

#define	ChType_TCP		1
#define	ChType_Listen		2
#define	ChType_PIPE		3
#define	ChType_STREAM		4
#define	ChType_KICK		5

#define	RF_Read			1
#define	RF_Execute		2
#define	RF_Write		4
#define	RF_Child		8
#define	RF_Query		16

#define	ServeGET		1
#define	ServeHEAD		2
#define	ServePOST		3



typedef struct {
	int				num_samp;
	int				max_samp, min_samp;
	int				total;
	int				num_samp_limit;
} Metric_t;

WEXT Metric_t metric_xfer, metric_time, metric_read, metric_write;
WEXT Metric_t metric_qrate, metric_chstack;
WEXT Metric_t metric_tprintf;
WEXT int	TotalReqs;

#define	BufMustVerify		0x0001
#define	BufCloseOnEmpty		0x0002

typedef struct {
	ulong_t				flags;
	uchar_t				*ptr;
	ulong_t				max;		// memory size of buffer
	ulong_t				base;		// index of first byte of buffer
	ulong_t				read;		// index of read pointer
	ulong_t				write;		// index of write pointer
	ulong_t				verify;		// index of verify pointer
} Buffer_t;

#define	_BufReadPtr( b )		((b)->ptr + ((b)->read - (b)->base))
#define	_BufWritePtr( b )	((b)->ptr + ((b)->write - (b)->base))
#define	_BufVerifyPtr( b )	((b)->ptr + ((b)->verify - (b)->base))
#define	_BufSize( b )		((b)->write - (b)->read)
#define	_BufFreeSpace( b )	((b)->max - ((b)->write - (b)->base))

#if 0
typedef struct {
	uchar_t				*buf;	
	int					buf_max;
	int					buf_len;
	int					buf_off;
} Buffer_t;
#endif

#define	CHEV_Read	1		// select says read
#define	CHEV_Write	2		// select says write
#define	CHEV_ReadE	3		// select says read error
#define	CHEV_Query	4		// read function triggered
#define	CHEV_Expire	5		// start terminating
#define	CHEV_Init	6		// initialize
#define	CHEV_Child	7		// child task terminated
#define	CHEV_Close	8		// channel closing
#define	CHEV_Alarm	9		// timed alarm



#define	MAX_other_fds	8
#define	MAX_other_ptr	16

typedef struct _Channel_ {
	int					fdr, fdw, fde;
	int					type;
	int					idx;
	int					expire, alarm;
	unsigned			RF;
	char					dname[64];
	
	Buffer_t			o, i, e;
	int (*chev_fp)(struct _Channel_ *, unsigned, int);

	struct	_Channel_	*SPtr, *ChildPtr;
// tpc
	int					addr_len;
	struct sockaddr_in	addr;

	void 				*STACK_p;
	unsigned			STACK_max;
	int				thread_pid;
	int					content_offset, content_length;
	int					serve_type;
	int					total_write;
	char				req_log[1024];
	int					req_status;
	int					total_time;
	int					create_time;
	int					access_time;
	char				*root;
	struct _VServer_	*VSPtr;
	struct _WService_	*wservptr;
	struct _VisitorTrack_	*VTPtr;
	int (*cxcb_fp)(struct _Channel_ *, int);
	char				**argv, **env;
	int					argc, enc, argcl, encl;
	int					other_fds[MAX_other_fds];
	void 					*other_ptr[MAX_other_ptr];
	char					*head_buffer;
} Channel_t;

#define	MaxChannels	64
WEXT int	NumChannels;
WEXT int	MaxNumChannels;
WEXT Channel_t	Channels[ MaxChannels ];

typedef int(ChEventCallback_t)(struct _Channel_ *, unsigned, int);

typedef int(ChildExitCallback_t)(Channel_t *, int);
typedef int(ServiceCallback_t)(Channel_t *, char *, char *, char *);
typedef int(BuiltinCallback_t)( int, char **, char **);

#define	WS_Valid	0x80000000ul
#define	WS_Thread	0x00010000ul
#define	WS_Inline	0x00020000ul
#define	WS_CGIQuery	0x01000000ul
#define	WS_CGIRoot	0x00100000ul
#define	WS_USERRoot	0x00200000ul

#define	WS_MatchMASK	0x0F00
#define	WS_MatchName	0x0100
#define	WS_MatchPrefix	0x0200
#define	WS_MatchExt	0x0400
#define	WS_MatchWild	(WS_MatchPrefix | WS_MatchExt)

#define	WS_PrioMASK	0x0F

typedef struct _WService_ {
	ulong_t			flags;
	char				*name;
	int				name_len;
	char				*ext;
	int				ext_len;
	char				*extra;
	char				*alias;
	char				*root;
	int				hit_count;
	Metric_t				metric_xfer, metric_time;
	char				*rule;
	struct _WService_	*next;
	ServiceCallback_t	*fp;
	BuiltinCallback_t	*xfp;
} WService_t;

typedef int(VServerInit_t)(struct _VServer_ *);

#define	VS_Valid	0x80000000ul
#define	VS_TCP		0x00000001ul
#define	VS_UDP		0x00000002ul
#define	VS_LOCALONLY	0x10000000ul

typedef struct _VServer_ {
	ulong_t		flags;
	char		*Address;
	char		*Port;
	char		*WWWRoot;
	char		*CGIRoot;
	char		*USERRoot;
	char		*LOGRoot;
	char		*LOGName;
	char		*ServerAdmin;
	char		*ServerName;
	char		*DefaultPage;
	int		ChildExpire;

	VServerInit_t	*vsinit_fp;
	
	int		NumWoddServices;
	int		MaxWoddServices;
	WService_t	*WoddServices;
	
	FILE			*logfp;
	int			log_counter;
	struct _VServer_	*log_slave;

	int				hit_count;
	Metric_t				metric_xfer, metric_time;

} VServer_t;

WEXT	VServer_t	*VServers[MAX_VSERVER];

#define EnableVisitorTrack	1

typedef struct _VisitorTrack_ {
	struct in_addr		iaddr;
	int					access_time;
	int					create_time;
	int					total_req;
	Metric_t				TperREQ;
	Metric_t				metric_xfer, metric_time;
} VisitorTrack_t;

#define	MaxVisitorTrack		256
WEXT	int NumVisitorTrack;
WEXT VisitorTrack_t *VisitorTrack;

#define	HHService	1
#define	HHBuiltin	2
#define	HHPolicy	3
#define	HHLocalOnly	4



// woddbuf.c

int ClearBuffer( Buffer_t *bptr );
int FreeBuffer( Buffer_t *bptr );
int _BufferResize( Buffer_t *bptr, int m );
int BufferEnqueue( Buffer_t *bptr, uchar_t *p, int n );
char *BufferMakeSpace( Buffer_t *bptr, int n );
int BufferFreeSpace( Buffer_t *bptr );
int BufferLength( Buffer_t *bptr );
char *BufferDequeue( Buffer_t *bptr, int n );
int BufferInject( Buffer_t *bptr, uchar_t *p, int n );
int BufferDelete( Buffer_t *bptr, int n );
int BufferMove( Buffer_t *dbptr, Buffer_t *sbptr );
int bprintf( Buffer_t *bptr, const char *frmt, ... );
int biprintf( Buffer_t *bptr, const char *frmt, ... );

// woddchannel.c

int ChannelEvent( Channel_t *ChPtr, unsigned chev, int stat );
Channel_t *ChannelCreate( Channel_t *PrChPtr, int type, ChEventCallback_t  *chev_fp );
int ChannelDistroy( Channel_t *ChPtr, int forced );
int ChannelAttachFD( Channel_t *ChPtr, int fd );
int ChannelRemoveFD( Channel_t *ChPtr, int fd );
int ChannelAttachPTR( Channel_t *ChPtr, void *ptr );
int ChannelRemovePTR( Channel_t *ChPtr, void *ptr );
Channel_t *ChannelIndex( int idx );


// woddsock.c

int ChannelWrite( Channel_t *ChPtr );
int ChannelRead( Channel_t *ChPtr );
int ChannelSelect();
int AttachListener();

// woddutil.c

#define	TestRF( cp, f )	((cp)->RF & (f))
#define	SetRF( ch, toset, toclear ) FL_SetRF( __FILE__, __LINE__, (ch), (toset), (toclear) )
#define	ClearRF( ch, toclear ) FL_SetRF( __FILE__, __LINE__, (ch), (0), (toclear) )
int FL_SetRF( char *file, int line, Channel_t *ChPtr, unsigned ToSet, unsigned ToClear );
char *RFtostr( unsigned v, char *prefstr );

int kill_thread_zombie( Channel_t *ChPtr );
int SetCurrentTime();
int GetCurrentTime();
int HexDump( char *p, int l );
char *tprintf( const char *frmt, ... );
char *dupprintf( const char *frmt, ... );
#ifndef __QNX__
int stricmp( char *a, char *b );
int strnicmp( char *a, char *b, int l );
#endif
int AddMetric( Metric_t *mp, int v );
int GetMetric( Metric_t *mp );
int GetMetricS( Metric_t *mp, int );


int RunWoddity();

// woddlog.o
#define	req_log_max	64
#define	err_log_max	64

WEXT char req_log_buffer[req_log_max][256];
WEXT int	req_log_head;
WEXT char err_log_buffer[err_log_max][256];
WEXT int	err_log_head;

int log_req_status( Channel_t *ChPtr );
int log_error( Channel_t *ChPtr, char *msg );


// woddhttpch.o
int AttachHTTP_Listen( VServer_t *VSPtr );
int ChEventHTTP_TCP( Channel_t *ChPtr, unsigned chev, int stat );
int ChEventHTTP_PIPE( Channel_t *ChPtr, unsigned chev, int stat );
int ChEventHTTP_Listen( Channel_t *ChPtr, unsigned chev, int stat );


// woddhttp.c

WService_t *AttachHTTPService( VServer_t *VSPtr, ServiceCallback_t *sfp, char *name, char *extra, ulong_t flags );
char *HTTP_status_str( int HTTP_status );
int InsertHTTPHeadder( Channel_t *ChPtr, int HTTP_status );
int MakeErrorPage( Channel_t *ChPtr, int HTTP_status );
int ServeHTTPRequest( Channel_t *ChPtr );
int HTTPAlias( Channel_t *ChPtr, char *flnm, char *opts, char *extra );

// woddhttpread.c

int HTTPError( Channel_t *ChPtr, char *flnm, char *opts, char *extra );
int HTTPRobots( Channel_t *ChPtr, char *flnm, char *opts, char *extra );
int HTTPReadFile( Channel_t *ChPtr, char *flnm, char *opts, char *content_type );

// woddhttpexec.c

int Thread_RunService( void *args[] );
WService_t *AttachHTTPBuiltin( VServer_t *, BuiltinCallback_t *xfp, char *name, char *extra, ulong_t flags );
int HTTPExec( Channel_t *ChPtr, char *flnm, char *opts, char *extra );
int HTTPBuiltin( Channel_t *ChPtr, char *flnm, char *opts, char *extra );

// woddhttpint.c

//int AttachHTTPInternal( int(*fp)(Buffer_t*,Buffer_t*,char*,int,char**,char**), char *name );
//int InitHTTPInternal();
//int HTTPInternal( Channel_t *ChPtr, char *flnm, char *opts, char *whatever );

// woddpagestatus.c

int MakeStatusPage( Channel_t *ChPtr, char *flnm, char *opts, char *extra );

int RegVisitorTrack( Channel_t *ChPtr, struct in_addr iaddr );


// woddconf.c
int ConfigWoddity( char *flnm );
int StandardConfigWoddity( char *port );
int InitVServers();


#ifdef __cplusplus
}
#endif

#endif
