/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define	__WoddityMem_Module__
#include "woddmem.h"

#include "woddity.h"

// extern int verbose;
// static int verbose=20;

typedef struct {
	char	alloc_sfile[16];
	int	alloc_sline;
	int	alloc_pid;
	int	alloc_size;
	char	free_sfile[16];
	int	free_sline;
	int	free_pid;
} memtrackdebug_t;

//.typedef struct memtrack_t
//.	This structure contains the following members ...
//.		void *p; - pointer to memory
//.		int	s; - size of allocation
//.		char *file; - source file of creation
//.		int	line; - line number of source file
//.	This structure is used to track memory allocations

typedef struct {
	unsigned long	avail;		// bit mapped availability of units
	unsigned long	locks;		// bit mapped block lock, disable free
	char 			*p;		// pointer to start
	int			s;			// total size
	int			blocks;		// size of units
	int			num;		// number of units in block
	memtrackdebug_t	*mtd_ptr;
} memtrack_t;

memtrack_t *memtrack;
int	num_memtrack;
int	Ktotal=0;
int	KMtotal=0;

//.memtrack_t *Get_memtrack()
//.	Get an empty memtrack_t
//.	Returns: on success, a pointer to an empty structure
//.		on failure, NULL

memtrack_t *Get_memtrack() {
	int i, n;
	memtrack_t *mtp;
	
	for (i=0; ; i++) {
		if (i>=num_memtrack) {
			n = (num_memtrack)?(num_memtrack<<1):64;
			mtp = calloc( n, sizeof( memtrack_t ) );
			if (memtrack) {
				memcpy( mtp, memtrack, num_memtrack * sizeof( memtrack_t ) );
			}
			free( memtrack );
			memtrack = mtp;
			num_memtrack = n;
		}
		if (memtrack[i].p == NULL) return &(memtrack[i]);
	}
	return NULL;
}

//.memtrack_t *Find_memtrack( void *p ) 
//.	Given a pointer to memory, find the memtrack_t that refers to it
//.	Returns: on success, a pointer to a memtrack_t 
//.		on failure, NULL

memtrack_t *Find_memtrack( char *p ) {
	int i, n;
	memtrack_t *mtp;
	
	for (i=0, mtp=memtrack; i<num_memtrack; i++, mtp++) {
		if (mtp->p == NULL) continue;
		if ((mtp->p <= p) && ((mtp->p+mtp->s) > p)) return mtp;
	}
	return NULL;
}

//.int Clean_memtrack()
//.	Free all memory malloc'd through kmem
//.	If the global var "verbose" isgreator than 4, this 
//.	function will display un-free'd blocks.

int Clean_memtrack() {
	int i;
	memtrack_t *mtp;
	
	if (memtrack == NULL) return 0;
	
	for (i=0, mtp=memtrack; i<num_memtrack; i++, mtp++) {
		if (mtp->p == NULL) continue;
		free( mtp->p );
		mtp->p = NULL;
		free( mtp->mtd_ptr );
		mtp->mtd_ptr = NULL;
	}
	if (verbose & 32) printf( "MEM-Cleanup: done %d\n", num_memtrack );
	free( memtrack );
	memtrack = NULL;
	num_memtrack = 0;
	return 0;
}

typedef struct {
	int			blocks;		// size of units
	int			num;		// total number of units 
	int			num_allocd;	// total number alloc'd
	int			alloc_size;	// total currently alloced
} WMHeuristic_t;

#define	MAX_WMHeuristic	16

int HTMLDump_memtrack( Buffer_t *bptr, int doall, char *BaseURL ) {
	WMHeuristic_t WMH[MAX_WMHeuristic];
	WMHeuristic_t *WMHp;
	int i, bi, fl, num_blocks, num_banks, num_alloc, hi, v;
	ulong_t	bb;
	memtrack_t *mtp;
	memtrackdebug_t *mtdp;
	
	if (memtrack == NULL) return 0;
	memset( WMH, 0, sizeof( WMH ) );
	
	for (i=0, mtp=memtrack, num_banks=num_blocks=num_alloc=0; i<num_memtrack; i++, mtp++) {
		if (mtp->p == NULL) continue;
		num_banks++;
		num_blocks+=mtp->num;
		for (bi=0, bb=1, mtdp=mtp->mtd_ptr; bi< mtp->num; bi++, bb<<=1, mtdp++ ) {
			if ((mtp->avail & bb) == 0) num_alloc++;
		}
	}

	bprintf( bptr, "General:  %d banks, %d/%d blocks<br>\n", num_banks, num_alloc, num_blocks );
	
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	for (i=0, mtp=memtrack; i<num_memtrack; i++, mtp++) {
		if (mtp->p == NULL) continue;
		fl = 1;

		for (hi=0, WMHp=WMH; hi<MAX_WMHeuristic; hi++, WMHp++) {
			if (WMHp->blocks == 0) {
				WMHp->blocks = mtp->blocks;
				break;
			}
			if (WMHp->blocks == mtp->blocks) break;
		}
		WMHp->num += mtp->num;
		
		bprintf( bptr, "<tr><td align=\"right\"><b>%d</b></td>", i  );
		bprintf( bptr, "<td align=\"right\"><b>%d</b></td><td align=\"right\"><b>%d</b></td>", mtp->num, mtp->blocks );
		bprintf( bptr, "<td align=\"right\"><b>%X</b></td>\n", mtp->avail );
//		bprintf( bptr, "</tr>\n" );
		
		for (bi=0, bb=1, mtdp=mtp->mtd_ptr; bi< mtp->num; bi++, bb<<=1, mtdp++ ) {
			if ((!doall) && (mtp->avail & bb)) continue;
			if (!mtdp->alloc_sfile[0]) continue;
			if (fl) {
				fl = 0;
			} else {
				bprintf( bptr, "<tr><td colspan=\"4\">&nbsp;</td>\n" );
			}
		
			bprintf( bptr, "<td align=\"right\"><b>%s</b></td><td align=\"right\">%d</td>",
				(mtp->avail & bb)?"*":((mtp->locks & bb)?"#":"&nbsp;"),
				bi );

			if (mtdp->alloc_pid == MainPID) bprintf( bptr, "<td align=\"center\">-</td>" );
			else bprintf( bptr, "<td align=\"right\">%d</td>", mtdp->alloc_pid );

			if (mtdp->alloc_sfile[0]) bprintf( bptr, "<td>%s:%d</td>", mtdp->alloc_sfile, mtdp->alloc_sline );
			else bprintf( bptr, "<td>&nbsp;</td>" );

			bprintf( bptr, "<td align=\"right\">%d</td>", mtdp->alloc_size );
			if (!(mtp->avail & bb)) {
				WMHp->alloc_size += mtdp->alloc_size;
				WMHp->num_allocd++;
			}
			
			if (mtdp->free_pid == MainPID) bprintf( bptr, "<td align=\"center\">-</td>" );
			else bprintf( bptr, "<td align=\"right\">%d</td>", mtdp->free_pid );

			if (mtdp->free_sfile[0]) bprintf( bptr, "<td>%s:%d</td>", mtdp->free_sfile, mtdp->free_sline );
			else bprintf( bptr, "<td>&nbsp;</td>" );
			
			if (BaseURL) bprintf( bptr, "<td><a href=\"%sm%d\">&nbsp;&gt;&gt;&nbsp;</a></td>\n", BaseURL, (i << 5) | bi );
			bprintf( bptr, "</tr>\n" );
			
		}
		if (fl) bprintf( bptr, "<td>-</td><td colspan=\"6\">&nbsp;</td></tr>\n" );
	}
	bprintf( bptr, "</table>\n" );
	
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>block size</td><td>available</td><td>in use</td><td>avr. size</td></tr>\n" );
	for (hi=0, WMHp=WMH; hi<MAX_WMHeuristic; hi++, WMHp++) {
		if (!WMHp->blocks) break;
		bprintf( bptr, "<tr>\n" );
		bprintf( bptr, "<td align=\"right\">%d</td>\n", WMHp->blocks );
		bprintf( bptr, "<td align=\"right\">%d</td>\n", WMHp->num );
		bprintf( bptr, "<td align=\"right\">%d</td>\n", WMHp->num_allocd );
		if (WMHp->num_allocd) bprintf( bptr, "<td align=\"right\">%d</td>", WMHp->alloc_size / WMHp->num_allocd );
		else bprintf( bptr, "<td align=\"center\">-</td>" );
		bprintf( bptr, "</tr>\n" );
	}
	bprintf( bptr, "</table>\n" );
	return 0;
}

int __mem_lock=0;
int __mem_next=0;

#include "signal.h"

TestMEMLock() {
	int	prev, mypid;
	
	mypid = getpid();
	
#if 1
	if (__mem_lock) {
		if (verbose) printf( "&" );
		sched_yield();
	}
#endif
	
	if (__mem_lock == 0) {
		__mem_lock = mypid;
		return;
	} 

	prev = __mem_next;
	__mem_next = mypid;
	
	while (__mem_lock) {
		if (verbose) printf( "*** MEMLock %d ***\n", mypid );
		pause();
//		sched_yield();
	}
	__mem_lock = mypid;
	__mem_next = prev;
}

ClearMEMLock() {
	int	prev, mypid;
	
	__mem_lock = 0;

	if (prev = __mem_next) {
		__mem_next = 0;
		kill( prev, SIGCONT );
	}
	return 0;
}

KickMEMLock( int thispid ) {
	if (__mem_lock == thispid) ClearMEMLock();
}

//.void *Kmalloc( unsigned s, char *file, int line )

void *Kmalloc( unsigned s, char *file, int line ) {
	memtrack_t *mtp;
	memtrackdebug_t *mtdp;
	char *p;
	unsigned	bs, bnum, bb;
	int i, bi, countsame=0, n;
	
	if (verbose & 32) {
		printf( "Kmalloc %d  file: %s:%d\n", s, file, line );
		fflush( stdout );
	}
	
	bnum = 32;
	
	if (s <= 128) bs = 128;			// 128 * 32 = 4096   a good minimum allocation unit
	else if (s <= 1024) { bs = 1024; bnum = 32; }
	else if (s <= 4096) { bs = 4096; bnum = 32; }
	else if (s <= 8192) { bs = 8192; bnum = 16; }
	else if (s <= 32768) { bs = 32768; bnum = 8; }
	else { 
		for (bs = 65536; bs < s; bs<<=2);
		bnum = 4; 
		if (bs >= 0x40000) {
			bnum=2;
			if (bs >= 0x100000) bnum=1;
		}
	}

	TestMEMLock();
	p = NULL;
	
	for (i=0, mtp=memtrack; i>=0; i++, mtp++) {
		if (i >= num_memtrack) {
			i = -99;		// force this to be the last loop
			
			countsame>>=1;
			if (countsame) bnum<<=countsame;
			if (bnum > 32) bnum = 32;

			mtp = Get_memtrack();
			mtp->blocks = bs;
			mtp->num = bnum;
			mtp->s = bs*bnum;
	
			if ((mtp->p = malloc( mtp->s )) == NULL) break;
			mtp->mtd_ptr = malloc( bnum * sizeof( memtrackdebug_t ) );
			
			KMtotal += mtp->s;

			switch (bnum) {
				case 32:	mtp->avail = 0xFFFFFFFFul;		break;
				case 16:	mtp->avail = 0xFFFFul;    		break;
				case 8: 	mtp->avail = 0xFFul;      		break;
				case 4: 	mtp->avail = 0xFul;      		break;
				case 2: 	mtp->avail = 0x3ul;      		break;
				case 1: 	mtp->avail = 0x1ul;      		break;
				default:
					for (bi=0, bb=1; bi< mtp->num; bi++, bb<<=1 ) {
						mtp->avail |= bb;
					}
			}
		}
		if (mtp->p == NULL) continue;
		if (mtp->blocks != bs) continue;
		if (mtp->avail == 0) {
			countsame++;
			continue;
		}
		for (bi=0, bb=1, p = mtp->p; bi< mtp->num; bi++, bb<<=1, p+=bs ) {
			if (mtp->avail & bb) {
				mtp->avail ^= bb;
				Ktotal += mtp->blocks;
				ClearMEMLock();
				if (mtp->mtd_ptr) {
					mtdp = (mtp->mtd_ptr)+bi;
					strncpy( mtdp->alloc_sfile, file, sizeof( mtdp->alloc_sfile ) );
					mtdp->alloc_sline = line;
					mtdp->alloc_pid = getpid();
					mtdp->alloc_size = s;
				}
				return p;
			}
		}
	}

	ClearMEMLock();
	fprintf( stderr, "*** Kmalloc failed  file: %s:%d req: %d (%d) ***\n", file, line, s, bs );
	exit( 1 );
	return NULL;
}

//.void *Kcalloc( unsigned s, unsigned n, char *file, int line )

void *Kcalloc( unsigned s, unsigned n, char *file, int line ) {
	void	*p;
	s*=n;
	if (verbose & 32) printf( "Kcalloc - " );
	if ((p = Kmalloc( s, file, line )) == NULL) return p;
	memset( p, 0, s );
	return p;
}

//.char *Kstrdup( char *s, char *file, int line )

char *Kstrdup( char *s, char *file, int line ) {
	int l;
	char *p;
	
	if (!s) return NULL;
	l = strlen( s );
	if (l <= 0) return NULL;
	if (verbose & 32) printf( "Kstrdup - " );
	p = Kmalloc( l+1, file, line );
	strcpy( p, s );
	return p;
}

//.char *Kstrperm( char *s, char *file, int line )

static char *Kstrperm_ptr;
static int Kstrperm_len;
#define Kstrperm_blocksize	1024

char *Kstrperm( char *s, char *file, int line ) {
	int l;
	char *p;
	
	if (!s) return NULL;
	l = strlen( s );
	if (l <= 0) return NULL;
	if ((Kstrperm_ptr == NULL) || ((Kstrperm_len + l + 1) >= Kstrperm_blocksize)) {
		if (verbose & 32) printf( "Kstrperm - " );
		Kstrperm_ptr = Kmalloc( Kstrperm_blocksize, file, line );
		Kstrperm_len = 1;	// prevent accedental freeing
		malloc_lock( Kstrperm_ptr, 1 ); 
	}
	p = Kstrperm_ptr + Kstrperm_len;
	Kstrperm_len += l + 1;
	strcpy( p, s );
	return p;
}

//.char *Kstrasn( char *s, char *file, int line )

char *Kstrasn( char **tsp, char *s, char *file, int line ) {
	int l;
	char *p;
	
	if (s) l = strlen( s );
	else l = 0;

	if (l <= 0) {
		if (tsp && *tsp) {
			free( *tsp );
			*tsp = NULL;
		}
		
		return NULL;
	}
	if (verbose & 32) printf( "Kstrasn - " );

	if (tsp && *tsp) {
		p = Krealloc( *tsp, l+1, file, line );
	} else {
		p = Kmalloc( l+1, file, line );
	}
	strcpy( p, s );
	if (tsp) *tsp = p;
	return p;
}

//.void Kfree( void *p )

void Kfree( void *p, char *file, int line  ) {
	memtrack_t *mtp;
	memtrackdebug_t *mtdp;
	char	*bp, *es;
	unsigned	bb;
	int		bi;
	
	if (verbose & 32) {
		printf( "Kfree %p  file: %s:%d\n", p, file, line );
		fflush( stdout );
	}
	if (!p) return;
	if ((mtp = Find_memtrack( p )) == NULL) {
		es = tprintf( "*** free error - unknown block %s:%d ***", file, line );
		log_error( NULL, es );
		if (verbose) printf( "%s\n", es );
		else fprintf( stderr, "%s\n", es );
		free( p );
	} else {
		if (verbose & 64) printf( "Kfree %p for %d (%d)\n", mtp->p, mtp->s, Ktotal );

		for (bi=0, bb=1, bp = mtp->p; bi< mtp->num; bi++, bb<<=1, bp+=mtp->blocks ) {
			if (bp == p) {
				if (mtp->locks & bb) {
					es = tprintf( "*** free error - memory locked %s:%d ***", file, line );
					log_error( NULL, es );
					if (verbose) printf( "%s\n", es );
					else fprintf( stderr, "%s\n", es );
				} else {
					mtp->avail |= bb;
					Ktotal -= mtp->blocks;
				}
				if (mtp->mtd_ptr) {
					mtdp = (mtp->mtd_ptr)+bi;
					strncpy( mtdp->free_sfile, file, sizeof( mtdp->free_sfile ) );
					mtdp->free_sline = line;
					mtdp->free_pid = getpid();
				}
				return;
			}
		}
		es = tprintf( "*** free error - block not found %s:%d ***", file, line );
		log_error( NULL, es );
		if (verbose) printf( "%s\n", es );
		else fprintf( stderr, "%s\n", es );
	}
	return;
}


//.void *Krealloc( void *p, unsigned s, char *file, int line )

void *Krealloc( void *p, unsigned s, char *file, int line ) {
	memtrack_t *mtp;
	void *np;
	
	if (p == NULL) {
		if (verbose & 32) printf( "Krealloc - " );
		return Kmalloc( s, file, line );
	}
	if ((mtp = Find_memtrack( p )) == NULL) {
		fprintf( stderr, "*** realloc error - unknown block ***\n" );
	}
	
	
	if (verbose & 32) printf( "Krealloc - " );
	if (s < mtp->blocks) return p;
	np = Kmalloc( s, file, line );
	if (s < mtp->blocks) s = mtp->blocks;
	memcpy( np, p, s );
	if (malloc_lock( p, 0 ) == 1) malloc_lock( np, 1 );
	Kfree( p, file, line );

	return np;
}

//.unsigned malloc_size( void *p )

unsigned malloc_size( void *p  ) {
	memtrack_t *mtp;
	memtrackdebug_t *mtdp;
	char	*bp;
	unsigned	bb;
	int		bi;
	
	if (!p) return 0;
	if ((mtp = Find_memtrack( p )) == NULL) return 0;
	
	for (bi=0, bb=1, bp = mtp->p; bi< mtp->num; bi++, bb<<=1, bp+=mtp->blocks ) {
		if (bp == p) {
			if (mtp->mtd_ptr) {
				mtdp = (mtp->mtd_ptr)+bi;
				mtdp->alloc_size = mtp->blocks;
			}
			break;
		}
	}
	return mtp->blocks;
}

//.unsigned malloc_lock( void *p, int st )

int malloc_lock( void *p, int st ) {
	memtrack_t *mtp;
	char	*bp;
	unsigned	bb;
	int		bi, pst;
	
	if (p && ((mtp = Find_memtrack( p )) != NULL)) {
		for (bi=0, bb=1, bp = mtp->p; bi< mtp->num; bi++, bb<<=1, bp+=mtp->blocks ) {
			if (bp == p) {
				pst = (mtp->locks & bb)?1:0;
				if (st) mtp->locks |= bb;
				else mtp->locks &= ~bb;
				return pst;
			}
		}
	}
	errno = EINVAL;
	return -1;
}

void *malloc_finder( int mi ) {
	memtrack_t *mtp;
	int	mti, bi;
	
	if (memtrack == NULL) return NULL;

	mti = mi >> 5;
	bi = mi & 0x1F;
	if (mti > num_memtrack) return NULL;
	mtp = memtrack+mti;
	if (mtp->p == NULL) return NULL;
	if (bi > mtp->num) return NULL;
	
	return (mtp->p) + (bi * mtp->blocks);
}
