/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"

int HTMLDump_memtrack( Buffer_t *bptr, int, char * );
static char *MSP_url();

extern int Ktotal;
extern int KMtotal;
extern VServer_t *VServers[];

// extern int NumWoddServices;
// extern int MaxWoddServices;
// extern WService_t *WoddServices;

#define	VT_none		0
#define	VT_channels	1
#define	VT_accesslog	2
#define	VT_errorlog	3
#define	VT_fdstat	4
#define	VT_memstat	5
#define	VT_memall	6
#define	VT_rules	7
#define	VT_VisitorTrack	8
#define	VT_VTandRules	9
#define	VT_ChannelFD	10
#define	VT_RawVisitorTrack	11
#define	VT_channelsV	12
#define	VT_MAX	12

#define	VT_bufinfo		101
#define	VT_memdump	102


static char *MSP_flnm="status";
static int MSP_refresh = 0;
static int MSP_viewtype = 0;
static int MSP_buftype;
static int MSP_bufindex;
static int MSP_bufoffset;
static char MSP_url_buf[256];

typedef struct {
	char	*id;
	int	v;
} MSP_IDMapping_t;

static MSP_IDMapping_t MSP_Viewtype_Map[] = {
#if _Woddity_Visitor_StatusPage
	"Visitors", 	VT_VisitorTrack,
#endif
#if _Woddity_Rules_StatusPage
	"Rules", 		VT_rules,
#endif
#if _Woddity_Channel_StatusPage
	"Channels", 		VT_channels,
#endif
#if _Woddity_FD_StatusPage
	"FD Tracking", 		VT_fdstat,
#endif
#if _Woddity_AccessLog_StatusPage
	"Access Log", 		VT_accesslog,
#endif
#if _Woddity_ErrorLog_StatusPage
	"Error Log", 		VT_errorlog,
#endif
#if _Woddity_MEM_StatusPage
	"Mem Tracking", 	VT_memstat,
	"Full Mem Track", 	VT_memall,
#endif
#if _Woddity_Visitor_StatusPage
	"Raw Visitors", 	VT_RawVisitorTrack,
#endif
#if _Woddity_Channel_StatusPage
	"Channels V", 		VT_channelsV,
#endif
	"none",			VT_none,
	NULL, 0, };
	
static MSP_IDMapping_t MSP_XViewtype_Map[] = {
	"Buffer Info", 		VT_bufinfo,
	"Memory Dump", 		VT_memdump,
	NULL, 0, };
	
static MSP_IDMapping_t MSP_Refresh_Map[] = {
	"manual",	0,
	"slow",		60,
	"medium",	10,
	"fast",			1,
	NULL,		0, };

#define MSP_IDM_CWID	6
	
int MSP_IDM_List( Buffer_t *bptr, MSP_IDMapping_t *IDMPtr, int cur_value, char *opt_char, char *label ) {
	int	i, c;
	bprintf( bptr, "<table border=\"0\" cellspacing=\"3\" cellpadding=\"2\" width=\"96%%\">\n" );
	for (i=0; IDMPtr->id; i++, IDMPtr++) {
		c = i % MSP_IDM_CWID;
		if (c == 0) bprintf( bptr, "<tr><td><nobr>%s</nobr></td>\n", (i)?"":label );
//		if (c == 0) bprintf( bptr, "<tr><td width=\"20%%\"><nobr>%s</nobr></td>\n", (i)?"":label );
//		bprintf( bptr, "<td bgcolor=\"%s\" align=\"center\"><nobr>", (cur_value == IDMPtr->v)?"#33FF99":"#339999" );
		bprintf( bptr, "<td bgcolor=\"%s\" align=\"center\">", (cur_value == IDMPtr->v)?"#33FF99":"#339999" );
		bprintf( bptr, "<a href=\"%s%s%d\">", MSP_url(), opt_char, IDMPtr->v );
		bprintf( bptr, "%s", IDMPtr->id );
		bprintf( bptr, "</a></td>\n" );
		if (c == (MSP_IDM_CWID - 1)) bprintf( bptr, "</tr>\n" );
	}
	if (c != (MSP_IDM_CWID-1)) bprintf( bptr, "</tr>\n" );
	bprintf( bptr, "</table>\n" );
	return i;
}

char *MSP_IDM_str( MSP_IDMapping_t *IDMPtr, int value ) {
	int	i;
	for (i=0; IDMPtr->id; i++, IDMPtr++) {
		if (IDMPtr->v == value) return IDMPtr->id;
	}
	return NULL;
}


static char *MSP_url() {
	char *s;
	
	s = MSP_url_buf;
	s += sprintf( s, "%s?", MSP_flnm );
	if (MSP_refresh) s+= sprintf( s, "d%d&", MSP_refresh );
	if (MSP_viewtype == VT_bufinfo) {
		if (MSP_buftype) s+= sprintf( s, "b%d%c&", MSP_bufindex, MSP_buftype );
		if (MSP_bufoffset>=0) s+= sprintf( s, "o%d&", MSP_bufoffset );
	} else if (MSP_viewtype == VT_memdump) {
		s+= sprintf( s, "m%d&", MSP_bufindex );
		if (MSP_bufoffset>=0) s+= sprintf( s, "o%d&", MSP_bufoffset );
	} else {
		if (MSP_viewtype) s+= sprintf( s, "t%d&", MSP_viewtype );
	}
	*s=0;
	return MSP_url_buf;
}

static int MSP_parse_args( char *opts ) {
	char *s, *t;
	char	opt;
	int v;
	
	for (s=opts; s && *s; s=t) {
		t=s;
		while (*t && (*t != '&')) t++;
		if (*t) *t++ = 0;
		while (*s) {
			opt = *s++;
			switch (opt) {
				case 'd':
					v = strtol( s, &s, 0 );
					if ((v < 0) || (v > 1000000)) break;
					switch (*s) {
						case 'd':	v *= 24;
						case 'h':	v *= 60;
						case 'm':	v *= 60;
						case 's':	break;
					}
					MSP_refresh = v;
					break;
				case 'b':
					v = strtol( s, &s, 0 );
					if ((v < 0) || (v > 1000000)) break;
					if ((MSP_buftype = *s) == 0) break;
					MSP_bufindex = v;
					MSP_viewtype = VT_bufinfo;
					break;
				case 'm':
					v = strtol( s, &s, 0 );
					if ((v < 0) || (v > 1000000)) break;
					MSP_bufindex = v;
					MSP_viewtype = VT_memdump;
					break;
				case 'o':
					v = strtol( s, &s, 0 );
					MSP_bufoffset = v;
					break;
				case 't':
					v = strtol( s, &s, 0 );
					if ((v < 0) || (v > VT_MAX)) break;
					MSP_viewtype = v;
					break;
			}
			if (islower( opt )) break;
		}
	}
	return 0;
}

#define	HD_Width	32
static int MSP_HexDump( Buffer_t *bptr, uchar_t *ptr, unsigned boff, unsigned bmax ) {
	unsigned	bo, bi, n, i, v;
	
	bprintf( bptr, "<table border=\"0\" cellpadding=\"1\" cellspacing=\"1\">\n" );
			
	for (n=0, bo=boff; (n<34) && (bo<bmax); n++, bo+=HD_Width) {
		bprintf( bptr, "<tr><td align=\"right\">%X&nbsp;</td>\n", bo );
		for (i=0, bi = bo; i<HD_Width; i++, bi++) {
			if ((bi < 0) || (bi >= bmax)) {
				bprintf( bptr, "<td><tt>--</tt></td>" );
				continue;
			}
			v = *(ptr + bi);
			if ((i & 7) == 0 ) bprintf( bptr, "<td></td>" );
			bprintf( bptr, "<td><tt>%02X</tt></td>", v & 0xFF );
		}
		bprintf( bptr, "<td>&nbsp;&nbsp;</td><td bgcolor=\"#FFFFFF\"><tt>" );
		for (i=0, bi = bo; i<HD_Width; i++, bi++) {
			if ((bi < 0) || (bi >= bmax)) v = '.';
			else v = (*(ptr + bi)) & 0xFF;
			
			switch (v) {
				case '<':	bprintf( bptr, "&lt;" );	break;
				case '>':	bprintf( bptr, "&gt;" );	break;
				default:
					if (isprint( v )) bprintf( bptr, "%c", v );
					else bprintf( bptr, "." );
			}
		}
		bprintf( bptr, "</td></tr>\n" );
				
	}			
	bprintf( bptr, "</table>\n" );
	return 0;
}

static int PrRelTime( Buffer_t *bptr, int v ) {
	int n;
	
	if (v == 0) {
		bprintf( bptr, "<td align=\"center\">-</td>"  );
		return 0;
	}
	if ((v = GetCurrentTime() - v) >= 60000) {
		n = v/1000;
		if (v < 3600000) {
			bprintf( bptr, "<td align=\"right\">%d:%02d.%03d</td>",  n/60, n%60, v%1000  );
		} else {
			bprintf( bptr, "<td align=\"right\">%d:%02d:%02d</td>", n/3600,  (n/60)%60, n%60  );
		}
	} else {
		bprintf( bptr, "<td align=\"right\">%d.%03d</td>", v/1000,  v%1000  );
	}
}

static int PrMetric( Buffer_t *bptr, Metric_t *mptr, char *units ) {
	if (mptr->num_samp) {
		bprintf( bptr, "<td>%d%s  (%d-%d)</td>\n", GetMetric( mptr ), units, mptr->min_samp, mptr->max_samp ); 
	} else {
		bprintf( bptr, "<td align=\"center\">-</td>\n" ); 
		
	}
}

#define	MAX_DispVisitorTrack	64
#define	MaxAccessTime		(5 * 60 * 1000)
typedef struct {
	int	val;
	VisitorTrack_t	*VTPtr;
} VTSort_t;

static void MSP_SortVisitorTrack( Buffer_t *bptr ) {
	VisitorTrack_t	*VTPtr, *qVTPtr;
	VTSort_t	*VTSBuf, *VTSPtr, *qVTSPtr, VTS_tmp;
	int	i, n, v, mq, mqr;
	int	nums;
	
	if (!VisitorTrack) return;

	VTSBuf = calloc( MAX_DispVisitorTrack, sizeof( VTSort_t ) );
	nums=0;
	
	for (i=0, VTPtr = VisitorTrack; i<NumVisitorTrack; i++, VTPtr++) {
		v = VTPtr->TperREQ.total + (GetCurrentTime() - VTPtr->access_time);
		mq = VTPtr->TperREQ.num_samp+1;
		if (mq > 0) v /= mq;
		else v = 0;
		
		if (nums < MAX_DispVisitorTrack) {
			qVTSPtr = VTSBuf+nums;
			nums++;
		} else {
			for (VTSPtr = VTSBuf, n=0, mq=v, qVTSPtr=NULL; n<nums; n++, VTSPtr++) {
				if (VTSPtr->val > mq) continue;
				mq = VTSPtr->val;
				qVTSPtr = VTSPtr;
			}
		}
		if (!qVTSPtr) continue;
		mq = VTPtr->TperREQ.total + v;
		qVTSPtr->val = v;
		qVTSPtr->VTPtr = VTPtr;
	}
#if 1
	for (i=nums-1; i>0; i--) {
		for (VTSPtr = VTSBuf, n=0; n<i; n++, VTSPtr++) {
			if (VTSPtr->val < (VTSPtr+1)->val) continue;
			VTS_tmp = *VTSPtr;
			*VTSPtr = *(VTSPtr+1);
			*(VTSPtr+1) = VTS_tmp;
		}
		
	}
#endif
	
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>addr</td><td>first</td><td>last</td>"
				"<td>hits</td><td>req/s</td><td>avr size</td><td>avr time</td>"
				"</tr>\n" );

//	for (i=0, VTPtr = VisitorTrack; i<NumVisitorTrack; i++, VTPtr++) {

	for (VTSPtr = VTSBuf, i=0; i<nums; i++, VTSPtr++) {
		VTPtr = VTSPtr->VTPtr;
		bprintf( bptr, "<tr>\n" );
		bprintf( bptr, "<td>%s</td>\n", inet_ntoa( VTPtr->iaddr ) );
		PrRelTime( bptr, VTPtr->create_time );
		PrRelTime( bptr, VTPtr->access_time );
		bprintf( bptr, "<td align=\"right\">%d</td>\n", VTPtr->total_req );

		if (VTPtr->total_req >= 4) {
			mq = (VTPtr->TperREQ.total + (GetCurrentTime() - VTPtr->access_time)) * 100;
			v = VTPtr->TperREQ.num_samp+1;
			if (v > 0) mq /= v;
			else mq = 0;
//			mq = GetMetricS( &(VTPtr->TperREQ), 100 );
			if (mq) mqr = 10000000 / mq;
			else mqr=0;
			bprintf( bptr, "<td>%d.%02d (%d.%02dms)</td>", mqr/100, mqr%100, mq/100, mq%100 ); 
		} else {
			bprintf( bptr, "<td>-</td>\n" );
		}
		PrMetric( bptr, &(VTPtr->metric_xfer), "B" );
		PrMetric( bptr, &(VTPtr->metric_time), "ms" );
//		bprintf( bptr, "<td></td>\n" );
		bprintf( bptr, "</tr>\n" );
	}
	bprintf( bptr, "</table>\n" );
	free( VTSBuf );
}

static void MSP_VisitorTrack( Buffer_t *bptr ) {
	VisitorTrack_t	*VTPtr;
	int	i, v, mq, mqr;
	
	if (!VisitorTrack) return;
	
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>addr</td><td>first</td><td>last</td>"
				"<td>hits</td><td>req/s</td><td>avr size</td><td>avr time</td>"
				"</tr>\n" );
	for (i=0, VTPtr = VisitorTrack; i<NumVisitorTrack; i++, VTPtr++) {
		bprintf( bptr, "<tr>\n" );
		bprintf( bptr, "<td>%s</td>\n", inet_ntoa( VTPtr->iaddr ) );
		PrRelTime( bptr, VTPtr->create_time );
		PrRelTime( bptr, VTPtr->access_time );
		bprintf( bptr, "<td align=\"right\">%d</td>\n", VTPtr->total_req );

		if (VTPtr->total_req >= 4) {
			mq = (VTPtr->TperREQ.total + (GetCurrentTime() - VTPtr->access_time)) * 100;
			v = VTPtr->TperREQ.num_samp+1;
			if (v > 0) mq /= v;
			else mq = 0;
//			mq = GetMetricS( &(VTPtr->TperREQ), 100 );
			if (mq) mqr = 10000000 / mq;
			else mqr=0;
			bprintf( bptr, "<td>%d.%02d (%d.%02dms)</td>", mqr/100, mqr%100, mq/100, mq%100 ); 
		} else {
			bprintf( bptr, "<td>-</td>\n" );
		}
		PrMetric( bptr, &(VTPtr->metric_xfer), "B" );
		PrMetric( bptr, &(VTPtr->metric_time), "ms" );
//		bprintf( bptr, "<td></td>\n" );
		bprintf( bptr, "</tr>\n" );
	}
	bprintf( bptr, "</table>\n" );
}

MSP_errorlog( Buffer_t *bptr ) {
	int i, n;
	bprintf( bptr, "<pre>\n" );
	for (i=err_log_head, n=1; n<err_log_max; n++) {
		if (--i < 0) i = err_log_max-1;
		if (!err_log_buffer[i][0]) break;
		bprintf( bptr, "%3d  %s", n, err_log_buffer[i] );
	}
	bprintf( bptr, "</pre>\n" );
}

MSP_accesslog( Buffer_t *bptr ) {
	int i, n;
	bprintf( bptr, "<pre>\n" );
	for (i=req_log_head, n=1; n<req_log_max; n++) {
		if (--i < 0) i = req_log_max-1;
		if (!req_log_buffer[i][0]) break;
		bprintf( bptr, "%3d  %s", n, req_log_buffer[i] );
	}
	bprintf( bptr, "</pre>\n" );
}

MSP_fdstat( Buffer_t *bptr ) {
	RegFD_t		*fdp;
	int i;

	if (!RegFD_ptr) return;
			
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>fd</td><td>channel</td><td>pid</td><td>flags</td><td>type</td>"
				"<td>access</td><td>open</td><td>close</td><td>staterr</td>"
				"<td>close stat</td><td>staterr type</td>"
				"<td>open ref</td><td>close ref</td>"
				"</tr>\n" );
						
	for (i=0, fdp = RegFD_ptr; i<RegFD_num; i++, fdp++) {
		bprintf( bptr, "<tr>" );
		bprintf( bptr, "<td align=\"right\">%d</td>", fdp->fd );
		bprintf( bptr, "<td align=\"right\">%d</td>", fdp->cidx );
				
		if (fdp->pid == MainPID) bprintf( bptr, "<td align=\"center\">-</td>" );
		else bprintf( bptr, "<td align=\"right\">%d</td>", fdp->pid );

		bprintf( bptr, "<td align=\"right\">%X</td>", fdp->flags );
		bprintf( bptr, "<td bgcolor=\"%s\">", (fdp->flags & RFD_fileopen)?"#99FF99":"#FF9999" );
//		bprintf( bptr, "%s ", (fdp->flags & RFD_fileopen)?"Open":"Closed" );
		if (fdp->flags & RFD_openFD) bprintf( bptr, "FD " );
		if (fdp->flags & RFD_openFILE) bprintf( bptr, "stream " );
		if (fdp->flags & RFD_openTCP) bprintf( bptr, "TCP " );
		if (fdp->flags & RFD_openPIPE) bprintf( bptr, "PIPE %c%c", 
			((fdp->flags & RFD_openPIPE1) || (fdp->flags & RFD_openPIPE1c))?'1':'0',
			((fdp->flags & RFD_openPIPE0c) || (fdp->flags & RFD_openPIPE1c))?' ':'x'  );

		bprintf( bptr, "</td>" );
			
		PrRelTime( bptr, fdp->access_tm );
		PrRelTime( bptr, fdp->open_tm );
		PrRelTime( bptr, fdp->close_tm );
		PrRelTime( bptr, fdp->staterr_tm );
		bprintf( bptr, "<td align=\"right\">%d</td>", fdp->close_stat );
		bprintf( bptr, "<td align=\"right\">%d</td>", fdp->staterr_type );
				
		if (*(fdp->open_sfile)) bprintf( bptr, "<td>%s:%d</td>", fdp->open_sfile, fdp->open_sline );
		else bprintf( bptr, "<td>&nbsp;</td>" );

		if (*(fdp->close_sfile)) bprintf( bptr, "<td>%s:%d</td>", fdp->close_sfile, fdp->close_sline );
		else bprintf( bptr, "<td>&nbsp;</td>" );
				
		bprintf( bptr, "</tr>\n" );
	}
			
	bprintf( bptr, "</table>\n" );
}

MSP_channels( Buffer_t *bptr, int lvl ) {
	Channel_t	*IChPtr;
	int		i;
	char *s;

	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>ch</td><td>type</td><td>state</td>"
				"<td>in</td><td>out</td><td>err</td>"
				"<td>family</td><td>t pid</td><td>acc time</td><td>written</td>"
				"<td>addr</td><td>req_stat</td><td>&nbsp;</td></tr>\n" );
					
	for (i=0, IChPtr = Channels; i<MaxChannels; i++, IChPtr++) {
		if ((lvl == 1) && (i >= NumChannels)) break;
		if ((lvl==0) && (!IChPtr->type)) continue;

		bprintf( bptr, "<tr><td>%d</td>", IChPtr->idx );
		bprintf( bptr, "<td bgcolor=\"%s\">", (IChPtr->type)?"#99FF99":"#FF9999" );
		switch (IChPtr->type) {
			case ChType_TCP:	bprintf( bptr, "TCP" ); 	break;
			case ChType_PIPE:	bprintf( bptr, "PIPE" ); 	break;
			case ChType_STREAM:	bprintf( bptr, "STREAM" ); 	break;
			case ChType_Listen:	bprintf( bptr, "Listen" ); 	break;
			case ChType_KICK:	bprintf( bptr, "Kick" ); 	break;
			default: 			bprintf( bptr, "%d", IChPtr->type );
		}
		bprintf( bptr, "</td><td>%s</td>", RFtostr( IChPtr->RF, NULL ) );

		bprintf( bptr, "<td>" );

		if (IChPtr->fdr) bprintf( bptr, "%d ", IChPtr->fdr );
		else bprintf( bptr, "- " );
		if (IChPtr->i.ptr) bprintf( bptr, "<a href=\"%sb%di\">(%d/%dK)</a>", MSP_url(), IChPtr->idx, BufferLength( &(IChPtr->i) ), IChPtr->i.max>>10 );

		bprintf( bptr, "</td><td>" );

		if (IChPtr->fdw) bprintf( bptr, "%d ", IChPtr->fdw );
		else bprintf( bptr, "- " );
		if (IChPtr->o.ptr) bprintf( bptr, "<a href=\"%sb%do\">(%d/%dK)</a>", MSP_url(), IChPtr->idx, BufferLength( &(IChPtr->o) ), IChPtr->o.max>>10 );

		bprintf( bptr, "</td><td>" );

		if (IChPtr->fde) bprintf( bptr, "%d ", IChPtr->fde );
		else bprintf( bptr, "- " );
		if (IChPtr->e.ptr) bprintf( bptr, "<a href=\"%sb%de\">(%d/%dK)</a>", MSP_url(), IChPtr->idx, BufferLength( &(IChPtr->e) ), IChPtr->e.max>>10 );

		bprintf( bptr, "</td>" );

		if (IChPtr->SPtr) bprintf( bptr, "<td>p=%d</td>", IChPtr->SPtr->idx );
		else if (IChPtr->ChildPtr) bprintf( bptr, "<td>c=%d</td>", IChPtr->ChildPtr->idx );
		else bprintf( bptr, "<td>&nbsp;</td>" );

		if (IChPtr->thread_pid) bprintf( bptr, "<td align=\"right\">%d</td>", IChPtr->thread_pid );
		else bprintf( bptr, "<td>&nbsp;</td>" );

		PrRelTime( bptr, IChPtr->access_time);

		if (IChPtr->total_write) bprintf( bptr, "<td align=\"right\">%d</td>", IChPtr->total_write );
		else bprintf( bptr, "<td>&nbsp;</td>" );

		if (IChPtr->addr_len) bprintf( bptr, "<td align=\"right\">%s:%d</td>", inet_ntoa( IChPtr->addr.sin_addr ), ntohs( IChPtr->addr.sin_port ) );
		else bprintf( bptr, "<td>&nbsp;</td>" );

		bprintf( bptr, "<td align=\"right\">%d</td>", IChPtr->req_status );
		bprintf( bptr, "<td>%s</td>", IChPtr->req_log );
		bprintf( bptr, "</tr>\n" );
	}
	bprintf( bptr, "</table>\n" );	
}

MSP_bufinfo( Buffer_t *bptr ) {
	Channel_t	*IChPtr;
	Buffer_t	*Ibptr;

	Ibptr = NULL;
	switch (MSP_buftype) {
		case 'i':
			if ((IChPtr = ChannelIndex( MSP_bufindex )) == NULL) break;
			Ibptr = &(IChPtr->i);
			break;
		case 'o':
			if ((IChPtr = ChannelIndex( MSP_bufindex )) == NULL) break;
			Ibptr = &(IChPtr->o);
			break;
		case 'e':
			if ((IChPtr = ChannelIndex( MSP_bufindex )) == NULL) break;
			Ibptr = &(IChPtr->e);
			break;
	}
	if (Ibptr == NULL) {
		bprintf( bptr, "<h2>Invalid buffer</h2>" );
		return;
	}
	if (MSP_bufoffset < 0) MSP_bufoffset = Ibptr->read - Ibptr->base;
			
	bprintf( bptr, "<table>\n" );
	bprintf( bptr, "<tr>\n" );
	bprintf( bptr, "<td align=\"right\">flags: %X</td>", Ibptr->flags );
	bprintf( bptr, "<td align=\"right\">ptr: %p</td>", Ibptr->ptr );
	bprintf( bptr, "<td align=\"right\">max: %X</td>", Ibptr->max );
	bprintf( bptr, "<td align=\"right\">base: %X</td>", Ibptr->base );
	bprintf( bptr, "<td align=\"right\">read: %X</td>", Ibptr->read );
	bprintf( bptr, "<td align=\"right\">write: %X</td>", Ibptr->write );
	bprintf( bptr, "<td align=\"right\">verify: %X</td>", Ibptr->verify );
	bprintf( bptr, "<tr>\n" );
	bprintf( bptr, "<tr>\n" );
	bprintf( bptr, "<td align=\"right\" colspan=\"3\">relative</td>" );
	bprintf( bptr, "<td align=\"right\"><a href=\"%so%d\">base: %X</a></td>", MSP_url(), 0, 0 );
	bprintf( bptr, "<td align=\"right\"><a href=\"%so%d\">read: %X</a></td>", MSP_url(), Ibptr->read-Ibptr->base, Ibptr->read-Ibptr->base );
	bprintf( bptr, "<td align=\"right\"><a href=\"%so%d\">write: %X</a></td>", MSP_url(), Ibptr->write-Ibptr->base, Ibptr->write-Ibptr->base );
	bprintf( bptr, "<td align=\"right\"><a href=\"%so%d\">verify: %X</a></td>", MSP_url(), Ibptr->verify-Ibptr->base, Ibptr->verify-Ibptr->base );
	bprintf( bptr, "<tr>\n" );
	bprintf( bptr, "</table>\n" );

	bprintf( bptr, "<table border=\"1\" cellpadding=\"4\" cellspacing=\"3\"><tr>\n" );
	bprintf( bptr, "<td><a href=\"%so%d\">-200</a></td>\n", MSP_url(), MSP_bufoffset-0x200 );
	bprintf( bptr, "<td><a href=\"%so%d\">-100</a></td>\n", MSP_url(), MSP_bufoffset-0x100 );
	bprintf( bptr, "<td><a href=\"%so%d\">-10</a></td>\n", MSP_url(), MSP_bufoffset-0x10 );
	bprintf( bptr, "<td><a href=\"%so%d\">-1</a></td>\n", MSP_url(), MSP_bufoffset-0x1 );
	bprintf( bptr, "<td>%X</td>\n", MSP_bufoffset );
	bprintf( bptr, "<td><a href=\"%so%d\">+1</a></td>\n", MSP_url(), MSP_bufoffset+0x1 );
	bprintf( bptr, "<td><a href=\"%so%d\">+10</a></td>\n", MSP_url(), MSP_bufoffset+0x10 );
	bprintf( bptr, "<td><a href=\"%so%d\">+100</a></td>\n", MSP_url(), MSP_bufoffset+0x100 );
	bprintf( bptr, "<td><a href=\"%so%d\">+200</a></td>\n", MSP_url(), MSP_bufoffset+0x200 );
	bprintf( bptr, "</table>\n" );
		
	MSP_HexDump( bptr, Ibptr->ptr, MSP_bufoffset, Ibptr->max );			
}

MSP_memdump( Buffer_t *bptr ) {
	void	*ptr;
	int mx;

	if ((ptr = malloc_finder( MSP_bufindex )) == NULL) {
		bprintf( bptr, "<h2>Invalid buffer</h2>" );
		return;
	}
	mx = malloc_size( ptr );
	if (MSP_bufoffset < 0) MSP_bufoffset = 0;

	bprintf( bptr, "<table border=\"1\" cellpadding=\"4\" cellspacing=\"3\"><tr>\n" );
	bprintf( bptr, "<td><a href=\"%so%d\">-200</a></td>\n", MSP_url(), MSP_bufoffset-0x200 );
	bprintf( bptr, "<td><a href=\"%so%d\">-100</a></td>\n", MSP_url(), MSP_bufoffset-0x100 );
	bprintf( bptr, "<td><a href=\"%so%d\">-10</a></td>\n", MSP_url(), MSP_bufoffset-0x10 );
	bprintf( bptr, "<td><a href=\"%so%d\">-1</a></td>\n", MSP_url(), MSP_bufoffset-0x1 );
	bprintf( bptr, "<td>%X - %X</td>\n", MSP_bufoffset, mx );
	bprintf( bptr, "<td><a href=\"%so%d\">+1</a></td>\n", MSP_url(), MSP_bufoffset+0x1 );
	bprintf( bptr, "<td><a href=\"%so%d\">+10</a></td>\n", MSP_url(), MSP_bufoffset+0x10 );
	bprintf( bptr, "<td><a href=\"%so%d\">+100</a></td>\n", MSP_url(), MSP_bufoffset+0x100 );
	bprintf( bptr, "<td><a href=\"%so%d\">+200</a></td>\n", MSP_url(), MSP_bufoffset+0x200 );
	bprintf( bptr, "</table>\n" );
			
	MSP_HexDump( bptr, ptr, MSP_bufoffset, mx );			
}


MSP_rules( Buffer_t *bptr ) {
	VServer_t	*VSPtr;
	WService_t	*wptr;
	int i, vi;
	char *s;

	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td></td><td>flags</td>"
				"<td>match</td><td>extra</td><td>rule</td>"
				"<td>hits</td><td>avr size</td><td>avr time</td>"
				"</tr>\n" );
							
	for (vi=0; vi<MAX_VSERVER; vi++) {
		if ((VSPtr = VServers[vi]) == NULL) break;
		bprintf( bptr, "<tr>" );
		bprintf( bptr, "<td colspan=\"5\"><b>" );
		
		bprintf( bptr, "%s:%s", VSPtr->ServerName, VSPtr->Port );
		if (VSPtr->flags & VS_LOCALONLY) bprintf( bptr, " <i>LocalOnly</i>" );
		
		bprintf( bptr, "</b></td>" );

		bprintf( bptr, "<td align=\"right\">%d</td>", VSPtr->hit_count );
		PrMetric( bptr, &(VSPtr->metric_xfer), "B" );
		PrMetric( bptr, &(VSPtr->metric_time), "ms" );

		bprintf( bptr, "</tr>" );
		

		for (i=0, wptr = VSPtr->WoddServices; i<VSPtr->NumWoddServices; i++, wptr++) {
			bprintf( bptr, "<tr>" );
			bprintf( bptr, "<td align=\"right\">%d</td>", i );
			bprintf( bptr, "<td>%s%s%s%d</td>", 
				(wptr->flags & WS_Thread)?"T":((wptr->flags & WS_Inline)?"I":"?"),
				(wptr->root)?"R":((wptr->flags & WS_CGIRoot)?"C":((wptr->flags & WS_USERRoot)?"U":"W")),
				(wptr->alias)?"a":"",
				(wptr->flags & WS_PrioMASK) );
			bprintf( bptr, "<td>" );
			if (wptr->flags & WS_MatchName) {
				bprintf( bptr, "%s", (wptr->name && *(wptr->name))?wptr->name:"-" );
			} else {
				if (wptr->flags & WS_MatchPrefix) {
					bprintf( bptr, "%s", (wptr->name && *(wptr->name))?wptr->name:"-" );
				}
				bprintf( bptr, "*" );
				if (wptr->flags & WS_MatchExt) {
					bprintf( bptr, "%s", (wptr->ext && *(wptr->ext))?wptr->ext:"-" );
				}
			}
			bprintf( bptr, "</td>" );
			
			bprintf( bptr, "<td>%s</td>", (wptr->extra && *(wptr->extra))?wptr->extra:"-" );
			bprintf( bptr, "<td>%s</td>", (wptr->rule && *(wptr->rule))?wptr->rule:"<i>internal</i>" );
			bprintf( bptr, "<td align=\"right\">%d</td>", wptr->hit_count );
			PrMetric( bptr, &(wptr->metric_xfer), "B" );
			PrMetric( bptr, &(wptr->metric_time), "ms" );
			bprintf( bptr, "</tr>\n" );
		}
	}
			
	bprintf( bptr, "</table>\n" );
}

MSP_systeminfo( Buffer_t *bptr, int lvl, int numchan, int numfd ) {
	int mx, mt, mq, mqr, kps, i, v, n, bo, bi;
	char	*s;
	void	*ptr;

	mx = GetMetric( &metric_xfer );
	mt = GetMetric( &metric_time );
	mq = GetMetricS( &metric_qrate, 100 );
	if (mq) mqr = 10000000 / mq;
	else mqr=0;
	
	bprintf( bptr, "<table border=\"1\" cellpadding=\"2\" cellspacing=\"0\">\n" );
	bprintf( bptr, "<tr><td>Requests / up-time</td><td>%d / \n", TotalReqs );
	
//	bprintf( bptr, "<tr><td>Up-time</td><td>" );
	v = GetCurrentTime() - StartTime + (WTMinute/3);
	if (n = RunDays + (v / WTDay)) bprintf( bptr, "%d days ", n );
	v = (v % WTDay) / WTMinute;
	bprintf( bptr, "%d:%02d", v/60, v%60 );
	bprintf( bptr, "</td></tr>\n" );

	if (mqr) {
		kps = ((mx * mqr)+50000) / 100000;
//		bprintf( bptr, "<tr><td>Request data rate</td><td>%d kB/s</td></tr>\n", kps ); 
	} else kps = 0;

	bprintf( bptr, "<tr><td>Req per sec</td><td>%d.%02d / %d.%02dms / %dkB/s</td></tr>\n", mqr/100, mqr%100, mq/100, mq%100, kps ); 

	bprintf( bptr, "<tr><td>Channels / Children / FD</td><td>%d (%d)\n", numchan, MaxNumChannels );
	bprintf( bptr, "&nbsp;/&nbsp;%d (%d)\n", TotalChildren, MaxTotalChildren );
	bprintf( bptr, "&nbsp;/&nbsp;%d (%d)</td></tr>\n", numfd, RegFD_maxFD );
	if (lvl>=1) bprintf( bptr, "<tr><td>Average xfer </td><td>%dB  (%d-%d)</td></tr>\n", mx, metric_xfer.min_samp, metric_xfer.max_samp ); 
	if (lvl>=1) bprintf( bptr, "<tr><td>Average time</td><td>%dms (%d-%d)</td></tr>\n", mt, metric_time.min_samp, metric_time.max_samp ); 
	bprintf( bptr, "<tr><td>In use memory</td><td>%dK (%d.%02dM / \n", Ktotal>>10, Ktotal>>20, ((Ktotal & 0xFFFFF) * 100) >> 20 ); 
	bprintf( bptr, " %d.%02dM)</td></tr>\n", KMtotal>>20, ((KMtotal & 0xFFFFF) * 100) >> 20 ); 
	if (lvl>=1) bprintf( bptr, "<tr><td>Thread Stack </td><td>%dB  (%d-%d)</td></tr>\n", GetMetric( &metric_chstack), metric_chstack.min_samp, metric_chstack.max_samp ); 
	if (lvl>=1) bprintf( bptr, "<tr><td>tprintf time</td><td>%dms (%d-%d)</td></tr>\n", GetMetric( &metric_tprintf ), metric_tprintf.min_samp, metric_tprintf.max_samp ); 

	
	bprintf( bptr, "</table>\n" );

}

int MakeStatusPage( Channel_t *ChPtr, char *flnm, char *opts, char *extra ) {
	int mx, mt, mq, mqr, kps, i, numchan, v, n, bo, bi, numfd;
	Buffer_t	*bptr;
	Channel_t	*IChPtr;
	RegFD_t		*fdp;
	char	*s;
	void	*ptr;
	
	MSP_refresh = 0;
	MSP_viewtype = 0;
	MSP_bufindex=-1;
	MSP_buftype=0;
	MSP_bufoffset=-1;
	
	MSP_flnm = flnm;
	if (opts) MSP_parse_args( opts );
	
	bptr = &(ChPtr->o);
	
	for (i=0, numchan=0, IChPtr = Channels; i<=NumChannels; i++, IChPtr++) {
		if (IChPtr->type == 0) continue;
		numchan++;
	}

	numfd=0;
	if (RegFD_ptr) {
		for (i=0, fdp = RegFD_ptr; i<RegFD_num; i++, fdp++) {
			if (fdp->flags & RFD_fileopen) numfd++;
		}
	}

	bprintf( bptr, "Content-type: text/html\n" );
	bprintf( bptr, "Expires: 0\n" );
	bprintf( bptr, "\n" );
	bprintf( bptr, "<html><head><title>%s - Status</title>\n", ServerSoftware );
	if (MSP_refresh) bprintf( bptr, "<META HTTP-EQUIV=\"Refresh\" CONTENT=\"%d\">\n", MSP_refresh );

	bprintf( bptr, 	"</head>\n"
				"<body bgcolor=\"#FFFFEE\" text=\"#333300\" link=\"#3300CC\" vlink=\"#0000CC\">\n"
				"<h1> %s - Status</h1>\n",
				ServerSoftware );


	bprintf( bptr, "<table border=\"0\"><tr valign=\"top\"><td>\n" );
	
	MSP_systeminfo( bptr, (MSP_viewtype==VT_none)?1:0, numchan, numfd );

	bprintf( bptr, "</td><td>&nbsp;</td><td>\n" );

	MSP_IDM_List( bptr, MSP_Viewtype_Map, MSP_viewtype, "t", "Report:" );
	
	MSP_IDM_List( bptr, MSP_Refresh_Map, MSP_refresh, "d", tprintf( "Refresh (%d): ", MSP_refresh ) );
	
	bprintf( bptr, "Test pages: " );
	bprintf( bptr, "<a href=\"index.html\">index.html</a>, " );
	bprintf( bptr, "<a href=\"index.rbml\">index.rbml</a>, " );
	bprintf( bptr, "<a href=\"csi\">lots of images</a>, " );
	bprintf( bptr, "<a href=\"csi?d10\">every 10 seconds</a>, " );
	bprintf( bptr, "<br>\n" );
	
	bprintf( bptr, "</td></tr></table><br>\n" );
	if ((s = MSP_IDM_str( MSP_Viewtype_Map, MSP_viewtype )) == NULL)
		if ((s = MSP_IDM_str( MSP_XViewtype_Map, MSP_viewtype )) == NULL) 
			s = "unknown";
		
	bprintf( bptr, "<h2>%s</h2>\n", s );
	
	
	switch (MSP_viewtype) {
		case VT_errorlog:	MSP_errorlog( bptr );
			break;
		case VT_accesslog:	MSP_accesslog( bptr );
			break;
		case VT_memstat:	
			HTMLDump_memtrack( bptr, 0, MSP_url() );
			break;
		case VT_memall:		
			HTMLDump_memtrack( bptr, 1, MSP_url() );
			break;
		case VT_fdstat:		MSP_fdstat( bptr );
			break;
		case VT_channels:	MSP_channels( bptr, 0 );
			break;
		case VT_channelsV:	MSP_channels( bptr, 1 );
			break;
		case VT_bufinfo:	MSP_bufinfo( bptr );
			break;
		case VT_memdump:	MSP_memdump( bptr );
			break;
		case VT_rules:		MSP_rules( bptr );
			break;
		case VT_VisitorTrack:		MSP_SortVisitorTrack( bptr );
			break;
		case VT_RawVisitorTrack:		MSP_VisitorTrack( bptr );
			break;
		case VT_VTandRules:
			bprintf( bptr, "<table border=\"0\" cellpadding=\"4\" cellspacing=\"2\"><tr valign=\"top\"><td>\n" );
			MSP_SortVisitorTrack( bptr );
			bprintf( bptr, "</td><td>\n" );
			MSP_rules( bptr );
			bprintf( bptr, "</td></tr></table>\n" );
			break;
		case VT_ChannelFD:
			bprintf( bptr, "<table border=\"0\" cellpadding=\"4\" cellspacing=\"2\"><tr valign=\"top\"><td>\n" );
			MSP_channels( bptr, 0 );
			bprintf( bptr, "</td><td>\n" );
			MSP_fdstat( bptr );
			bprintf( bptr, "</td></tr></table>\n" );
			break;
	}

			
	bprintf( bptr, "</body></html>\n" );	

	ChPtr->req_status = 200;

	InsertHTTPHeadder( ChPtr, 200 );

	SetRF( ChPtr, RF_Write, RF_Execute );
	return 0;
}


