/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"
#include <sys/time.h>

int TestFD( Channel_t *ChPtr, int fd ) {
	struct stat statbuf;
	int	r;
	
	r = fstat( fd, &statbuf );	
	if (r == 0) return 0;
//	if (verbose & 2) printf( "*** FD Error on %d - fd %d\n", ChPtr->idx, fd );
	log_error( ChPtr, tprintf( "*** FD Error on %d - fd %d\n", ChPtr->idx, fd ) );
	RegFD( ChPtr->fdr, RFD_staterr, errno );
	return -1;
}

int ChannelSelect() {
	fd_set	readmask, writemask;
	int i, n, h, r, readnum, writenum, errnum, LastChannel=0;
	int max_fd;
	struct timeval	tv;
	Channel_t	*ChPtr;
	Buffer_t		*bptr;
	int	PrevSelectErr=0;
	
	if(verbose) printf( "Ready.\n" );
	
	SendKICK( "Start" );
	
	for (;;) {
		for (i=0, ChPtr = Channels; i<=NumChannels; i++, ChPtr++) {
			if (ChPtr->type == 0) continue;
			if (ChPtr->expire) {
				if ((n = (GetCurrentTime() - ChPtr->access_time)) > (ChPtr->expire)) {
					log_error( ChPtr, tprintf( "EXPIRE at %ds", n/1000 ) );
					ChannelEvent( ChPtr, CHEV_Expire, n );
				}
			}
			if (ChPtr->alarm) {
				if ((n = (GetCurrentTime() - ChPtr->alarm)) >= 0) {
					ChPtr->alarm = 0;
					ChannelEvent( ChPtr, CHEV_Alarm, n );
				}
			}
			kill_thread_zombie( ChPtr );
			if (ChPtr->RF == 0) {
				ChannelDistroy( ChPtr, 0 );
				continue;
			}
			if (ChPtr->type == 0) continue;
			if (TestRF( ChPtr, RF_Query)) {
				ClearRF( ChPtr, RF_Query);
				ChannelEvent( ChPtr, CHEV_Query, 0 );
			}
		}
		
		InSelect=1;

		RF_Changed = 0;
		max_fd = 0;
		readnum = 0;
		errnum = 0;
		writenum = 0;
		FD_ZERO( &readmask );
		FD_ZERO( &writemask );
		for (i=0, ChPtr = Channels; i<=NumChannels; i++, ChPtr++) {
			if (ChPtr->type == 0) continue;
			if (verbose&2) printf( "%d:%s%s%s%s%s%s ", ChPtr->idx, 
				(ChPtr->RF==0)?"-":"",
				(ChPtr->RF&RF_Read)?"r":"",
				(ChPtr->RF&RF_Execute)?"x":"",
				(ChPtr->RF&RF_Write)?"w":"",
				(ChPtr->RF&RF_Child)?"c":"",
				(ChPtr->RF&RF_Query)?"q":""
				);
			if ((verbose&2) && PrevSelectErr) {
				if (ChPtr->fdr) printf( "fdr:%d ", ChPtr->fdr );
				if (ChPtr->fdw) printf( "fdw:%d ", ChPtr->fdw );
				if (ChPtr->fde) printf( "fde:%d ", ChPtr->fde );
			}
			if (ChPtr->RF == 0) {
				ChannelDistroy( ChPtr, 0 );
				continue;
			}
			if (ChPtr->type == 0) continue;


			if (TestRF( ChPtr, RF_Read ) && (h = ChPtr->fdr)) {
				if (PrevSelectErr && TestFD( ChPtr, h )) {
					if (ChPtr->fdw == ChPtr->fdr) {
						ChPtr->fdw = 0;
						ClearRF( ChPtr, RF_Write );
					}
					RFDclose( ChPtr->fdr );
					ChPtr->fdr = 0;
					ClearRF( ChPtr, RF_Read );
				} else {
					FD_SET( h, &readmask );
					if (h >= max_fd) max_fd = h+1;
					if (ChPtr->type != ChType_KICK) readnum++;
				}
			}
			if (h = ChPtr->fde) {
				if (PrevSelectErr && TestFD( ChPtr, h )) {
					RFDclose( ChPtr->fde );
					ChPtr->fde = 0;
				} else {
					FD_SET( h, &readmask );
					if (h >= max_fd) max_fd = h+1;
					errnum++;
				}
			}
			if (TestRF( ChPtr, RF_Write ) && ChPtr->fdw) {
				if (PrevSelectErr && TestFD( ChPtr, ChPtr->fdw )) {
					RFDclose( ChPtr->fdw );
					ChPtr->fdw = 0;
					ClearRF( ChPtr, RF_Write );
				} else {

					if (ChPtr->SPtr) bptr = &(ChPtr->SPtr->i);
					else bptr = &(ChPtr->o);
				
					if (BufferLength( bptr ) > 0) {
						if (h = ChPtr->fdw) {
							FD_SET( h, &writemask );
							if (h >= max_fd) max_fd = h+1;
							writenum++;
						}
					} else {
						if (ChPtr->fdw != ChPtr->fdr) {
							RFDclose( ChPtr->fdw );
							ChPtr->fdw = 0;
							ClearRF( ChPtr, RF_Write );
						}
						if (!TestRF( ChPtr, RF_Execute)) ClearRF( ChPtr, RF_Write );
					}
				}
			}
		}
		if (verbose&2) fflush( stdout );
		if (verbose&8) printf( "Select for read %d, write %d, max %d, limit %d\n", 
			readnum, writenum, max_fd, FD_SETSIZE );

		if (RF_Changed) {
			InSelect=0;
			if (verbose&4) printf( "RFCHG:%X ", RF_Changed );
			continue;
		}

		if (readnum == 0) return 0;
		
		tv.tv_sec=60;		// drop out of select after 1 minute for house-keeping
		tv.tv_usec=0;
		InSelect=1;
		r = select( max_fd, &readmask, &writemask, NULL, &tv );
		InSelect=0;

		SetCurrentTime();

		if (r <= 0) {
			if ((r == 0) || (errno == EINTR)) {
				PrevSelectErr = 0;
				if (r == 0) {	// nothing to do
					ChannelHousekeeping();
				}
				continue;
			}
			log_error( NULL, tprintf( "Select ERROR %d was %d  (%dr %dw %de)", errno, PrevSelectErr, readnum, writenum, errnum ) );
			if (errno == PrevSelectErr) exit( 1 );
			PrevSelectErr = errno;
			continue;
		}
		PrevSelectErr = 0;
		
		for (i=0, n=0; n<NumChannels; n++, i++) {
			if (i >= NumChannels) i=0;
			ChPtr = Channels+i;
			if (ChPtr->type == 0) continue;
			if (h = ChPtr->fdr) {
				if (FD_ISSET( h, &readmask )) {
					FD_CLR( h, &readmask );
					if (verbose&2) printf( "%d=r ", ChPtr->idx );
					ChannelEvent( ChPtr, CHEV_Read, 0 );
					LastChannel = i;
				}
			}
			if (h = ChPtr->fde) {
				if (FD_ISSET( h, &readmask )) {
					FD_CLR( h, &readmask );
					if (verbose&2) printf( "%d=e ", ChPtr->idx );
					ChannelEvent( ChPtr, CHEV_ReadE, 0 );
					LastChannel = i;
				}
			}
		} 
		for (i=0, n=0; n<NumChannels; n++, i++) {
			if (i >= NumChannels) i=0;
			ChPtr = Channels+i;
			if (ChPtr->type == 0) continue;
			if (h = ChPtr->fdw) {
				if (FD_ISSET( h, &writemask )) {
					FD_CLR( h, &writemask );
					if (verbose&2) printf( "%d=w ", ChPtr->idx );
					ChannelEvent( ChPtr, CHEV_Write, 0 );
					LastChannel = i;
				}
			}
		} // end write testing
	}
}

#ifdef __QNX__
pid_t _select_receive( pid_t proxy ) {
	pid_t	pid;
	while (1) {
		pid = Receive( 0, 0, 0 );
//		SetCurrentTime();
		if (pid == proxy) return pid;
		if (pid == kick_proxy) return -1;
	}
}
#endif

int SendKICK( char *s ) {
	if (!kick_fd) AttachKICK();
	write( kick_fd, s, strlen( s ) );
	return 0;
}

int ChEventKICK( Channel_t *ChPtr, unsigned chev, int stat ) {
	int 		n, v;
	uchar_t		ibuf[256];

	if (!ChPtr->alarm) ChPtr->alarm = GetCurrentTime() + WTHour;
	if (!StartTime) StartTime = SetCurrentTime();

	switch (chev) {
		case CHEV_Read:
			n = read( ChPtr->fdr, ibuf, sizeof( ibuf ));
			ibuf[n] = 0;
			if (verbose&2) printf( "Kick read %d %s\n", n, ibuf );
			if (n > 64) strcpy( &ibuf[64], "..." );
			sprintf( ChPtr->req_log, "%s (%d) ", ibuf, n );
			break;
		case CHEV_Expire:
			ChannelDistroy( ChPtr, 0 );
			AttachKICK();
			break;
		case CHEV_Alarm:
			if ((n = GetCurrentTime() - StartTime) > WTDay) {
				v = n / WTDay;
				RunDays += v;
				StartTime += v * WTDay;
			}
			break;
		case CHEV_Close:
			RFDclose( kick_fd );
			kick_fd = 0;
			break;
		case CHEV_Init: {
			int pipeK[2];
	
			pipe( pipeK );
			ChPtr->fdr = pipeK[0];
			kick_fd = pipeK[1];
			SetRF( ChPtr, RF_Read, 0 );
			sprintf( ChPtr->req_log, "kick me quick %d", kick_fd );

			RegFD( ChPtr->fdr, RFD_open | RFD_openPIPE0c, ChPtr->idx );
			RegFD( kick_fd, RFD_open | RFD_openPIPE1, ChPtr->idx );

//			ChPtr->expire = 24*60*60*1000;
			FreeBuffer( &(ChPtr->i) );
			FreeBuffer( &(ChPtr->o) );
			FreeBuffer( &(ChPtr->e) );
			
			break;
		}
		default:	return -1;
	}
	return 0;
}

int AttachKICK() {
	ChannelCreate( NULL, ChType_KICK, ChEventKICK );
	return 0;
}

