/*--------------------------------------------------------------------------
    woddity - WebOddity, Web Server                              
    Copyright (C) 2001 - Robin Burgener, robinb@20q.net                     
                                                                            
    This program is free software; you can redistribute it and/or modify    
    it under the terms of the GNU General Public License as published by    
    the Free Software Foundation; either version 2 of the License, or       
    (at your option) any later version.                                     
                                                                            
    This program is distributed in the hope that it will be useful,         
    but WITHOUT ANY WARRANTY; without even the implied warranty of          
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
    GNU General Public License for more details.                            
                                                                            
    You should have received a copy of the GNU General Public License       
    along with this program; if not, write to the Free Software             
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
                                                                            
--------------------------------------------------------------------------*/

#include "woddity.h"
#ifdef __QNX__
#include "sys/sched.h"
#else
#include "sched.h"
#endif

char *RFtostr( unsigned v, char *prefstr ) {
	if (!v) {
		if (prefstr) return tprintf( "%s<nil>", prefstr );
		return "<nil>";
	}
	return tprintf( "%s%s%s%s%s%s",
			(prefstr)?prefstr:"",
			(v & RF_Read)?"r":"",
			(v & RF_Execute)?"x":"",
			(v & RF_Write)?"w":"",
			(v & RF_Child)?"c":"",
			(v & RF_Query)?"q":"" );
}

#define	SetRF( ch, toset, toclear ) FL_SetRF( __FILE__, __LINE__, (ch), (toset), (toclear) )

int FL_SetRF( char *file, int line, Channel_t *ChPtr, unsigned ToSet, unsigned ToClear ) {
	unsigned	p = ChPtr->RF;
	unsigned	c, s, a;
	
	if (ToClear) ChPtr->RF &= ~ToClear;
	ChPtr->RF |= ToSet;

	a = ChPtr->RF;	
	c = (ToClear & p) & ~a;	
	s = (ToSet & ~p) & a;	
	RF_Changed |= (c | s);
	if (c | s) {
		if (verbose & 4) printf( "%s:%d  Ch %d RF%s%s%s%s\n",
			file, line, 
			ChPtr->idx,
			RFtostr(p, " "),
			(s)?RFtostr(s, " +"):"",
			(c)?RFtostr(c, " -"):"",
			RFtostr(a, " =") );
	}
	return ChPtr->RF;
}

void KickMEMLock( int thispid );

int kill_thread_zombie( Channel_t *ChPtr ) {
	int stat=0, r;
	Channel_t	*PChPtr;
	if (ChPtr->thread_pid == 0) return 0;
	
	r = waitpid( ChPtr->thread_pid, &stat, WNOHANG );
	if (r < 0) {
//		if (errno = EINTR) return 0;
		return 0;
	} else if (r != ChPtr->thread_pid) return 0;
	
	KickMEMLock( ChPtr->thread_pid );
	
	if (ChPtr->STACK_p) {
		int i;
		ulong_t	*p;
		p = ChPtr->STACK_p;
		for (i=ChPtr->STACK_max; i>0; i-=4, p++) if (*p) break;
		AddMetric( &metric_chstack, i );
	}
	ChPtr->thread_pid = 0;
	TotalChildren--;
	
	ChannelEvent( ChPtr, CHEV_Child, stat );

	if (verbose & 2) printf( "kill_thread_zombie on %d stat %d (%d)\n", ChPtr->idx, stat, r );
	if (r < 0) perror( "kill_thread_zombie" );
	return 0;
}

// SetCurrentTime and GetCurrentTime
// SetCurrentTime is called once per select 
//	where as 
// GetCurrentTime should be used as the current time value 
// both functions return 32-bit time in milliseconds

#ifdef __QNX__
#include <sys/timers.h>
#include <sys/osinfo.h>

static time_t far *CurTimeSecPtr;
static long	far *CurTimeNSecPtr;
static int		CurrentTime;

int SetCurrentTime() {
	GetCurrentTime();
	if (verbose & 4) printf( "[%03d]", ((CurrentTime&0x7FFFFFFF)/100)%1000 );
	return CurrentTime;
}

int GetCurrentTime() {
	if (CurrentTime == 0) {
		struct	_osinfo	osdata;

		if (qnx_osinfo( (nid_t)0, &osdata) == -1) return( -1 );
		CurTimeSecPtr = MK_FP( osdata.timesel, 4 ); // 0 + offsetof ( struct _timesel, seconds) );
		CurTimeNSecPtr = MK_FP( osdata.timesel, 0 ); // 0 + offsetof ( struct _timesel, seconds) );
	}
	CurrentTime = ((*CurTimeSecPtr) * 1000) + ((*CurTimeNSecPtr)/1000000);
	if (!CurrentTime) CurrentTime++;
	return CurrentTime;
}
#else
#include <sys/timeb.h>

static int		CurrentTime;

int SetCurrentTime() {
	struct timeb tmb;
	ftime( &tmb );
	
	CurrentTime = (tmb.time * 1000) + tmb.millitm;
	if (!CurrentTime) CurrentTime++;
	if (verbose & 4) printf( "[%03d]", ((CurrentTime&0x7FFFFFFF)/100)%1000 );

	return CurrentTime;
}

int GetCurrentTime() {
	return CurrentTime;
}
#endif

int HexDump( char *p, int l ) {
	int i, v;
	char	cbuf[20];
	
	for (i=0; i<l; i++) {
		if ((i & 15) == 0) printf( "%04X ", i );
		v = *(p+i);
		printf( "%02X ", v&0xFF );
		cbuf[i&15] = ((v>=' ') && (v<0x7f))?v:'.';
		cbuf[(i&15)+1] = 0;
		if ((i & 15) == 15) printf( "  %s\n", cbuf );
	}
	if ((i & 15) != 0) printf( "  %s\n", cbuf );
	return 0;
}

char *dupprintf( const char *frmt, ... ) {
	va_list	arglist;
	int l;
	char tbuf[1024];
	
	va_start( arglist, frmt );
	
	l = vsnprintf( tbuf, sizeof( tbuf ), frmt, arglist );
//	if (verbose & 32) puts( tbuf );
	return strdup( tbuf );
}


#define	_tprintf_buf_len	256
#define	_tprintf_buf_max	256
int	_tprintf_buf_index=0;

char *_tprintf_buf;
int	_tprintf_utimes[_tprintf_buf_max];

char *tprintf( const char *frmt, ... ) {
	va_list	arglist;
	int l, bi, t;
	char *s;

	if (++_tprintf_buf_index >= _tprintf_buf_max) _tprintf_buf_index = 1;
	bi = _tprintf_buf_index;
	if (!_tprintf_buf) {
		if ((_tprintf_buf = calloc( _tprintf_buf_len, _tprintf_buf_max )) == NULL) {
			fprintf( stderr, "*** tprintf failed ***\n" );
			return NULL;
		}
		malloc_lock( _tprintf_buf, 1 );
	}
		
	if (_tprintf_utimes[bi]) {
		t = GetCurrentTime() - _tprintf_utimes[bi];
		AddMetric( &metric_tprintf, t );
	}
	_tprintf_utimes[bi] = GetCurrentTime();

	s = _tprintf_buf + (bi * _tprintf_buf_len);
		
	va_start( arglist, frmt );
	
	l = vsnprintf( s, _tprintf_buf_len-1, frmt, arglist );
	*(s+l)=0;
	return s;
}

int _vprintf( char *file, int line, int vlvl, const char *frmt, ... ) {
	va_list	arglist;
	char tbuf[1024];
	int l, bi;

	if ((vlvl & verbose) == 0) return 0;
		
	va_start( arglist, frmt );
	
	l = vsnprintf( tbuf, sizeof( tbuf ), frmt, arglist );
	puts( tbuf );
	return l;
}

#ifndef __QNX__
//.int stricmp( char *a, char *b )
//.	Compare 2 strings, case insensitive
//.	Returns: 0-match or !0-no match
int stricmp( char *a, char *b ) {
	while (*a && *b && (toupper( *a ) == toupper( *b ))) { a++; b++; }
	return *a - *b;
}

//.int strnicmp( char *a, char *b, int l )
//.	Compare 2 strings, upto "l" characters long, case insensitive
//.	Returns: 0-match or !0-no match
int strnicmp( char *a, char *b, int l ) {
	while ((l>0) && *a && *b && (toupper( *a ) == toupper( *b ))) { a++; b++; l--; }
	if (!l) return 0;
	return *a - *b;
}
#endif


int AddMetric( Metric_t *mp, int v ) {
	if (!mp->num_samp_limit) mp->num_samp_limit=1000;
	if ( (mp->num_samp >= mp->num_samp_limit) || 
			((mp->num_samp>0) && ((mp->total + v) > 0x400000)) ) {
		mp->total+=v-(mp->total / mp->num_samp);
	} else {
		mp->total += v;
		mp->num_samp++;
	}
	if ((mp->min_samp == 0) || (v < mp->min_samp)) mp->min_samp = v;
	if ((mp->max_samp == 0) || (v > mp->max_samp)) mp->max_samp = v;
	return 0;
}


int GetMetric( Metric_t *mp ) {
	if (mp->num_samp > 0) {
		return (mp->total / mp->num_samp);
	} 
	return 0;
}

int GetMetricS( Metric_t *mp, int S ) {
	if (mp->num_samp > 0) {
		return ((mp->total*S) / mp->num_samp);
	} 
	return 0;
}

unsigned long MakeHCode( char *s ) {
	unsigned long r=0x55555555;
	unsigned long	v;
	
	while (s && *s) {
		v = *s++;
		r += (r >> 5) + (r >> 19) + (r * 153) + (v * 737);
	}
	if (!r) r++;
	return r;
}

void HandleSIG( int signum ) {
	if (verbose & 4) printf( "*** %d:SIG %d ***\n", getpid(), signum );
}

static int _SIG_INT_COUNTER=0;
void HandleSIGexit( int signum ) {
	Channel_t	*ChPtr;
	int			i;
	
//	verbose |= 0xFF;
	
	if (verbose & 4) printf( "*** %d:SIGexit %d ***\n", getpid(), signum );

	for (i=0, ChPtr = Channels; i<=NumChannels; i++, ChPtr++) {
		if (ChPtr->type == 0) continue;
		if (ChPtr->thread_pid == 0) continue;
		kill_thread_zombie( ChPtr );
		ChannelDistroy( ChPtr, 1 );
	}
//	if (++_SIG_INT_COUNTER > 12) exit( 0 );
	
	exit( 0 );
}

void HandleSIGPIPE( int signum ) {
	if (verbose & 4) printf( "*** %d:SIGPIPE ***\n", getpid() );
}

void HandleSIGCHLD( int signum ) {
	Channel_t	*ChPtr;
	int			i;
	
	if (verbose & 4) printf( "*** %d:SIGCHLD ***\n", getpid() );

#if 0
	for (i=0, ChPtr = Channels; i<=NumChannels; i++, ChPtr++) {
		if (ChPtr->type == 0) continue;
		if (ChPtr->thread_pid == 0) continue;
		if (verbose & 4) printf( "*** SIGCHLD test %d ***\n", ChPtr->idx );
		kill_thread_zombie( ChPtr );
		if ((ChPtr->thread_pid == 0) && (ChPtr->RF == 0)) ChannelDistroy( ChPtr, 0 );
	}
#endif
	SendKICK( "SIGCHLD " );
	if (verbose & 4) printf( "*** DONE - SIGCHLD ***\n", getpid() );
}

void ExitCloseLogs( void ) {
	if (MainPID == getpid()) {
		fprintf( stderr, "Woddity EXIT\n" );
		log_error( NULL, NULL );
		log_req_status( NULL );
	}
}

int ThreadChannelSelect() {
	int	pid, m;
	char *STACK_p;
	unsigned	STACK_max;

	STACK_max = STACK_SIZE;
	STACK_p = malloc( STACK_max );
	if ((m = malloc_size( STACK_p )) > STACK_max) STACK_max = m;
	malloc_lock( STACK_p, 1 );
	
#ifdef __QNX__
	pid = _beginthread( Thread_RunService, STACK_p, STACK_max, NULL );
#else
	pid = __clone( ChannelSelect, STACK_p+STACK_max-64,
		SIGCHLD | CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND , NULL );
#endif	
	if (pid == -1) return -1;
	if (verbose&2) printf( "*** clone %d\n", pid );
	return 0;
}

int RunWoddity( int SparkThread ) {
	int i;
	
	ServerSoftware = strperm( tprintf( "WebOddity/%d.%02d%s", WODD_VER_MAJ, WODD_VER_MIN, WODD_VER_REV ) );
	
	atexit( ExitCloseLogs );
	
	srand( SetCurrentTime() );
	MainPID = getpid();
#if 0
	if (chdir( WWWRoot )) {
		perror( "Invalid WWWRoot" );
		exit( 1 );
	}
#endif

	for (i=1; i<32; i++) {
		switch (i) {
			case SIGSEGV:	break;
			case SIGCHLD:	break;
//			case SIGCHLD:	signal( i, HandleSIGCHLD );	break;
			case SIGPIPE:	signal( SIGPIPE, HandleSIGPIPE );	break;
			case SIGINT:	signal( SIGINT, HandleSIGexit );	break;
			default:	signal( i, HandleSIG );
		}
	}

	InitVServers();
//	AttachHTTP_Listen( NULL );
	
#ifdef __QNX__
	kick_proxy = qnx_proxy_attach( 0, NULL, 0, 0 );	
#endif
	if (SparkThread) return ThreadChannelSelect();
	else return ChannelSelect();
}


